/*******************************************************************************
*                         Goggles Music Manager                                *
********************************************************************************
*           Copyright (C) 2007-2008 by Sander Jansen. All Rights Reserved      *
*                               ---                                            *
* This program is free software: you can redistribute it and/or modify         *
* it under the terms of the GNU General Public License as published by         *
* the Free Software Foundation, either version 3 of the License, or            *
* (at your option) any later version.                                          *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program.  If not, see http://www.gnu.org/licenses.           *
********************************************************************************/
#include "common.h"
#include "md5.h"
#include "GMAudioScrobbler.h"


/* Needed for sockets */
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>


/* Offline Queue */
#define SCROBBLER_CACHE_FILE PATHSEPSTRING ".goggles" PATHSEPSTRING "scrobbler.cache" 

/* Where to submit tracks */
#define LASTFM_HOSTNAME "post.audioscrobbler.com"
#define LASTFM_PORT 80

#define MAX_HANDSHAKE_TIMEOUT 7200 

/* Submit every 1 minute */
#define SCHEDULE_DELAY 60

#define CLIENT_ID "gmm"
#define CLIENT_VERSION "0.1"


static void checksum(FXString & io){
  if (io.empty()) return;
  md5_state_t pms;
  md5_byte_t digest[16];

  md5_init(&pms);
  md5_append(&pms,(const md5_byte_t*)io.text(),io.length());
  md5_finish(&pms,digest);

  io.length(32);
#if FOXVERSION < FXVERSION(1,7,0)
  for (FXint i=0,d=0;i<32;i+=2,d++) {
    io[i]=FXString::hex[(digest[d]/16)%16];
    io[i+1]=FXString::hex[digest[d]%16];
    }
#else
  for (FXint i=0,d=0;i<32;i+=2,d++) {
    io[i]=FXString::value2Digit[(digest[d]/16)%16];
    io[i+1]=FXString::value2Digit[digest[d]%16];
    }
#endif

  }


static FXbool parse_url(const FXString & url,FXString & hostname,FXint & port,FXString & path){
  if (compare(url,"http://",7)==0){
    FXint p=url.find('/',7);
    FXint s=url.find(':',7);
    if (s==-1 || (p>=0 && s>p) ) {
      hostname=url.mid(7,p-7);
      port=80;
      path=url.right(url.length()-p);
      } 
    else {
      hostname=url.mid(7,s-7);
#if FOXVERSION < FXVERSION(1,7,0)      
      port=FXUIntVal(url.mid(s+1,p-s));
#else
      port=url.mid(s+1,p-s).toUInt();
#endif
      path=url.right(url.length()-p);
      }
    FXTRACE((70,"hostname=%s\nport=%d\npath=%s\n",hostname.text(),port,path.text()));
    return true;
    }
  return false;  
  }



void GMAudioScrobblerTrack::save(FXStream & store) const {
  store << artist;
  store << album;
  store << title;
  store << duration;
  store << no;
  store << timestamp;
  }

void GMAudioScrobblerTrack::load(FXStream & store) {
  store >> artist;
  store >> album;
  store >> title;
  store >> duration;
  store >> no;
  store >> timestamp;
  }




FXDEFMAP(GMAudioScrobbler) GMAudioScrobblerMap[]={
  FXMAPFUNC(SEL_CHORE,GMAudioScrobbler::ID_NEXT_TASK,GMAudioScrobbler::onNextTask),
  FXMAPFUNC(SEL_TIMEOUT,GMAudioScrobbler::ID_NEXT_TASK,GMAudioScrobbler::onNextTask),
  FXMAPFUNC(SEL_IO_READ,GMAudioScrobbler::ID_HANDSHAKE,GMAudioScrobbler::onHandshake),
  FXMAPFUNC(SEL_IO_WRITE,GMAudioScrobbler::ID_HANDSHAKE,GMAudioScrobbler::onHandshake),
  FXMAPFUNC(SEL_IO_EXCEPT,GMAudioScrobbler::ID_HANDSHAKE,GMAudioScrobbler::onHandshake),
  FXMAPFUNC(SEL_IO_READ,GMAudioScrobbler::ID_SUBMIT,GMAudioScrobbler::onSubmit),
  FXMAPFUNC(SEL_IO_WRITE,GMAudioScrobbler::ID_SUBMIT,GMAudioScrobbler::onSubmit),
  FXMAPFUNC(SEL_IO_EXCEPT,GMAudioScrobbler::ID_SUBMIT,GMAudioScrobbler::onSubmit),
  FXMAPFUNC(SEL_IO_READ,GMAudioScrobbler::ID_NOWPLAYING,GMAudioScrobbler::onNowPlaying),
  FXMAPFUNC(SEL_IO_WRITE,GMAudioScrobbler::ID_NOWPLAYING,GMAudioScrobbler::onNowPlaying),
  FXMAPFUNC(SEL_IO_EXCEPT,GMAudioScrobbler::ID_NOWPLAYING,GMAudioScrobbler::onNowPlaying)
  };

FXIMPLEMENT(GMAudioScrobbler,FXObject,GMAudioScrobblerMap,ARRAYNUMBER(GMAudioScrobblerMap))

/******************************************************************************
 *
 *  P U B L I C  A P I
 *
 ******************************************************************************/

GMAudioScrobbler::GMAudioScrobbler() {
  }

GMAudioScrobbler::GMAudioScrobbler(FXApp * a) : 
  app(a), 
  state(STATE_DISCONNECTED),
  disconnect(DISCONNECT_NONE), 
  server(-1),
  handshake_failed_timeout(60),
  failed(0),
  try_handshake(false),
  submit_npayload(0) {
  username=app->reg().readStringEntry("LastFM","username",NULL);
  password=app->reg().readStringEntry("LastFM","password",NULL);

  /// Check if we're banned
  if (compare(app->reg().readStringEntry("LastFM","client-id",CLIENT_ID),CLIENT_ID)==0 && 
      compare(app->reg().readStringEntry("LastFM","client-version",CLIENT_VERSION),CLIENT_VERSION)==0){
 
    if (app->reg().readBoolEntry("LastFM","client-banned",false))
      state=STATE_BANNED;
    }
  else { /// using different client, reset banned flag
    app->reg().writeBoolEntry("LastFM","client-banned",false);
    }
    
  app->reg().writeStringEntry("LastFM","client-id",CLIENT_ID);
  app->reg().writeStringEntry("LastFM","client-version",CLIENT_VERSION);
  load_queue();
  }

GMAudioScrobbler::~GMAudioScrobbler(){

  for (FXint e=dnscache.first();e<dnscache.size();e=dnscache.next(e)){
    freeaddrinfo((struct addrinfo*)dnscache.data(e));
    }

  app->removeChore(this,ID_NEXT_TASK);
  app->removeTimeout(this,ID_NEXT_TASK);     
  close_connection();
  save_queue();
  }


void GMAudioScrobbler::shutdown() {
  if (state==STATE_CONNECTING) {
    close_connection();
    }
  else if (state==STATE_CONNECTED){ 
    disconnect=DISCONNECT_SHUTDOWN;
    app->runWhileEvents();
    }
  app->removeChore(this,ID_NEXT_TASK);
  app->removeTimeout(this,ID_NEXT_TASK);     
  close_connection();
  save_queue();
  }

void GMAudioScrobbler::submit(FXlong timestamp,GMTrack & info){
  FXTRACE((60,"%s::submit\n",getClassName()));
  if (isEnabled() && info.time>30) {
    queue.append(GMAudioScrobblerTrack(timestamp,info));
    reschedule_submit();
    }
  }

void GMAudioScrobbler::nowplaying(GMTrack & info){
  if (isEnabled()) {
    if (state==STATE_BANNED)  ;
    else if (state==STATE_BADAUTH) goto x;
    else if (state==STATE_CONNECTING || state==STATE_CONNECTED) goto x;
    if (username.empty() || password.empty()) goto x;
    npt=GMAudioScrobblerTrack(1,info);
    app->removeTimeout(this,ID_NEXT_TASK);
    app->addChore(this,ID_NEXT_TASK);
    }
  return;
x:
  app->removeTimeout(this,ID_NEXT_TASK);
  app->removeChore(this,ID_NEXT_TASK);
  return;
  }

void GMAudioScrobbler::setLogin(const FXString & user,const FXString & pass) {
  FXTRACE((60,"%s::setLogin\n",getClassName()));
  FXString newpass=pass;
  checksum(newpass);
  if (user!=username || newpass!=password) {
    username=user;
    password=newpass;
    app->reg().writeStringEntry("LastFM","username",username.text());
    app->reg().writeStringEntry("LastFM","password",password.text());
    session.clear();
    switch(state) {
      case STATE_CONNECTING:
        close_connection();
        app->addChore(this,ID_NEXT_TASK);
        break;
      case STATE_CONNECTED: 
        disconnect=DISCONNECT_LOGIN;
        break;
      case STATE_BADAUTH:
        if (!username.empty() && !password.empty()){ 
          state=STATE_DISCONNECTED;
          try_handshake=true;
          reschedule_submit();
          }
        break;
      default: try_handshake=true; reschedule_submit(); break;     
      }
    }
  }


const FXchar * GMAudioScrobbler::getStatus(FXColor & clr) const {
  clr=FXRGB(0,255,0);
  if (!session.empty()) return "Ok";
  clr=FXRGB(0,0,0);
  switch(state) {
    case STATE_CONNECTING:
    case STATE_CONNECTED: 
      clr=FXRGB(0,0,0);
      return "Connecting...";
      break;
    case STATE_BADAUTH:
      clr=FXRGB(255,0,0);
      return "Login Failed!";
      break; 
    case STATE_BANNED:
      clr=FXRGB(255,0,0);
      return "Unsupported. Please upgrade Goggles Music Manager.";
      break; 
    default: return "Disconnected";
    }
  return "Disconnected";
  }

void GMAudioScrobbler::clear() {
  FXTRACE((60,"%s::clear\n",getClassName()));
  queue.clear();
  FXString filename = FXSystem::getHomeDirectory() + SCROBBLER_CACHE_FILE;
  FXFile::remove(filename);
  }


FXint GMAudioScrobbler::getQueueLength() const {
  return queue.no();
  }

FXbool GMAudioScrobbler::isBanned() const {
  return (state==STATE_BANNED);
  }

FXbool GMAudioScrobbler::isLoginFailed() const {
  return (state==STATE_BADAUTH);
  }


FXbool GMAudioScrobbler::isEnabled() const {
  return (state!=STATE_BANNED && !username.empty() && !password.empty());
  }

/*
    PRIVATE API
-------------------------------------------------------------------------------------------------
*/






#define URL_UNSAFE   "#$-_.+!*'><()\\,%\""          // Always Encode
#define URL_RESERVED ";/?:@=&"              // Only encode if not used as reserved by scheme


// Encode url string
FXString myencode(const FXString& url){
  register FXint p=0;
  register FXint c;
  FXString result;
  while(p<url.length()){
    c=url[p++];
//    if(!Ascii::isAlphaNumeric(c) && ((c<=' ' || c>='{') || strchr(URL_UNSAFE URL_RESERVED,c))){
    if (!Ascii::isAlphaNumeric(c)){
      result.append('%');
#if FOXVERSION < FXVERSION(1,7,0)
      result.append(FXString::HEX[(c>>4)&15]);
      result.append(FXString::HEX[c&15]);
#else
      result.append(FXString::value2Digit[(c>>4)&15]);
      result.append(FXString::value2Digit[c&15]);
#endif
      continue;
      }
    result.append(c);
    }
  return result;
  }


void GMAudioScrobbler::load_queue(){
  FXTRACE((60,"%s::load_queue\n",getClassName()));
  FXuint version,size;
  FXString filename = FXSystem::getHomeDirectory() + SCROBBLER_CACHE_FILE;
  FXFileStream store;
  if (store.open(filename,FXStreamLoad)){
    store >> version;
    if (version==20080501) {
      store >> size;  
      queue.no(size);
      for (FXint i=0;i<queue.no();i++){
        queue[i].load(store);
        }
      }
    store.close();
    if (queue.no()) reschedule_submit();
    }
  FXFile::remove(filename);
  }


void GMAudioScrobbler::save_queue(){
  FXTRACE((60,"%s::save_queue\n",getClassName()));
  FXuint version=20080501;
  FXString filename = FXSystem::getHomeDirectory() + SCROBBLER_CACHE_FILE;
  if (queue.no()) { 
    FXFileStream store;
    if (store.open(filename,FXStreamSave)){
      store << version;
      store << queue.no();
      for (FXint i=0;i<queue.no();i++)
        queue[i].save(store);
      }
    }
  else {
    FXFile::remove(filename);
    }
  } 



void GMAudioScrobbler::set_permanent_failure(FXuint s) {
  FXTRACE((60,"%s::set_permanent_failure\n",getClassName()));
  close_connection();
  clear_submit_payload();
  state=s;
  if (state==STATE_BANNED) {
    app->reg().writeBoolEntry("LastFM","client-banned",true);
    FXMessageBox::error(app,MBOX_OK,"LastFM Permanent Failure","This version of Goggles Music Manager is not supported by Last-FM.\nPlease upgrade to a newer version");
    }
  app->removeTimeout(this,ID_NEXT_TASK);
  app->removeChore(this,ID_NEXT_TASK);
  }

void GMAudioScrobbler::set_failed() {
  FXTRACE((60,"%s::set_failed\n",getClassName()));
  failed++;
  if (failed==3) {
    session.clear();
    }
  }

void GMAudioScrobbler::reschedule_submit() {
  FXTRACE((60,"%s::reschedule_submit\n",getClassName()));
  if (state==STATE_BANNED) goto x;
  else if (state==STATE_BADAUTH) goto x;
  else if (state==STATE_CONNECTING || state==STATE_CONNECTED) goto x;
  if (username.empty() || password.empty()) goto x;
  if (queue.no()==0 && !try_handshake) goto x;
  if (try_handshake) {
    app->removeTimeout(this,ID_NEXT_TASK);
    app->addChore(this,ID_NEXT_TASK);
    }
  else {
    if (!app->hasTimeout(this,ID_NEXT_TASK) && !app->hasChore(this,ID_NEXT_TASK)){
      app->addTimeout(this,ID_NEXT_TASK,TIME_SEC(SCHEDULE_DELAY));     
      }
    }
  return;
x:
  app->removeTimeout(this,ID_NEXT_TASK);
  app->removeChore(this,ID_NEXT_TASK);
  return;
  }



void GMAudioScrobbler::close_connection(){
  FXTRACE((60,"%s::close_connection\n",getClassName()));
  if (server!=-1) {
    app->removeInput(server,INPUT_READ|INPUT_EXCEPT|INPUT_WRITE);
    ::close(server);
    server=-1;
    }

  if (state!=STATE_BADAUTH && state!=STATE_BANNED)
    state=STATE_DISCONNECTED;

  if (disconnect==DISCONNECT_LOGIN)
    app->addChore(this,ID_NEXT_TASK);

  disconnect=DISCONNECT_NONE;
  }



static FXint tryConnect(struct addrinfo * entry) {
  FXint server = socket(entry->ai_family,entry->ai_socktype,entry->ai_protocol);
  if (server==-1) return -1;

  FXint opts=fcntl(server,F_GETFL);
  if (opts>=0) fcntl(server,F_SETFL,opts|O_NONBLOCK);
    
  FXint result = connect(server,entry->ai_addr,entry->ai_addrlen);
  if (result!=-1 || (result==-1 && errno==EINPROGRESS))
      return server;

  ::close(server);
  return -1;
  }

void GMAudioScrobbler::open_connection(const FXString &hostname,FXint port,FXuint message) {
  FXTRACE((60,"%s::open_connection\n",getClassName()));
  FXString lookupname = hostname+GMStringVal(port);
  struct addrinfo   hints;
  struct addrinfo * available=NULL,*entry=NULL;
  
  memset(&hints,0,sizeof(struct addrinfo));
  hints.ai_family=AF_UNSPEC;
  hints.ai_socktype=SOCK_STREAM;  

  if (server!=-1)
    return;

  available = (struct addrinfo *) dnscache.find(lookupname.text());
  if (available) {
    for (entry=available;entry;entry=entry->ai_next){
      server = tryConnect(entry);
      if (server!=-1) break;
      }
    if (server==-1) {
      dnscache.remove(lookupname.text());
      freeaddrinfo(available);
      }
    }

  if (server==-1) {
    if (getaddrinfo(hostname.text(),GMStringVal(port).text(),&hints,&available)!=0){
      return;
      }
    for (entry=available;entry;entry=entry->ai_next){
      server = tryConnect(entry);
      if (server!=-1) break;
      }
    if (server!=-1) {
      dnscache.insert(lookupname.text(),available);
      }
    else {
      freeaddrinfo(available);
      }
    }

  if (server!=-1) {
    state=STATE_CONNECTING;
#if FOXVERSION < FXVERSION(1,7,0)
    app->addInput(server,INPUT_READ|INPUT_WRITE|INPUT_EXCEPT,this,message);
#else
    app->addInput(this,message,server,INPUT_READ|INPUT_WRITE|INPUT_EXCEPT);
#endif
    }
  }

void GMAudioScrobbler::start_handshake() {
  FXTRACE((60,"%s::start_handshake\n",getClassName()));
  failed=0;
  try_handshake=false;
  open_connection(LASTFM_HOSTNAME,LASTFM_PORT,ID_HANDSHAKE);
  }

void GMAudioScrobbler::start_submit() {
  FXTRACE((60,"%s::start_submit\n",getClassName()));
  FXString hostname;
  FXString path;
  FXint port;
  if (parse_url(submit_url,hostname,port,path)) {
    create_submit_payload();
    open_connection(hostname,port,ID_SUBMIT);
    } 
  }

void GMAudioScrobbler::start_nowplaying() {
  FXTRACE((60,"%s::start_nowplaying\n",getClassName()));
  FXString hostname;
  FXString path;
  FXint port;
  if (parse_url(nowplaying_url,hostname,port,path)) {
    open_connection(hostname,port,ID_NOWPLAYING);
    } 
  }




long GMAudioScrobbler::onNextTask(FXObject*,FXSelector,void*){
  FXTRACE((60,"%s::onNextTask\n",getClassName()));
  if (queue.no() || try_handshake || npt.timestamp) {
    if (session.empty())
      start_handshake();
    else if (queue.no())
      start_submit();
    else if (npt.timestamp)
      start_nowplaying();
    }
  return 0;
  }


/******************************************************************************
 *
 *  S U B  M I S S I O N
 *
 ******************************************************************************/
void GMAudioScrobbler::create_submit_request(FXString & msg) {
  FXTRACE((60,"%s::create_submit_request\n",getClassName()));
  FXString hostname;
  FXString path;
  FXint port;
  parse_url(submit_url,hostname,port,path);
  msg="POST " + path + " HTTP/1.1\r\nHOST:" + hostname + "\r\nContent-Type: application/x-www-form-urlencoded\r\n";
  msg+=GMStringFormat("Content-Length: %d\r\n",submit_payload.length());
  msg+="\r\n";
  msg+=submit_payload;
  }

void GMAudioScrobbler::create_submit_payload() {
  FXTRACE((60,"%s::create_submit_payload\n",getClassName()));
  submit_npayload=0;
  submit_payload.clear();
  submit_payload+="s=";
  submit_payload+=myencode(session);
  for (FXint i=0;i<queue.no() && i<50;i++) {
    submit_payload+=GMStringFormat("&a[%d]=%s",i,myencode(queue[i].artist).text());
    submit_payload+=GMStringFormat("&t[%d]=%s",i,myencode(queue[i].title).text());
    submit_payload+=GMStringFormat("&i[%d]=%u",i,(FXuint)(queue[i].timestamp/1000000000));
    submit_payload+=GMStringFormat("&o[%d]=P",i);
    submit_payload+=GMStringFormat("&r[%d]",i);
    submit_payload+=GMStringFormat("&l[%d]=%d",i,queue[i].duration);
    submit_payload+=GMStringFormat("&b[%d]=%s",i,myencode(queue[i].album).text());
    submit_payload+=GMStringFormat("&n[%d]=%d",i,queue[i].no);
    submit_payload+=GMStringFormat("&m[%d]",i);
    submit_npayload++;
    }
  FXTRACE((71,"Payload: %s\n\n",submit_payload.text()));
  }

void GMAudioScrobbler::clear_submit_payload() {
  submit_npayload=0;
  submit_payload.clear();
  }

void GMAudioScrobbler::process_submit_response(const FXchar * buffer,FXint len){
  FXTRACE((60,"%s::process_submit_response\n",getClassName()));
  if (len>0 && compare("HTTP/1.1 200",buffer,12)==0) { 
    FXStringList headers;
    FXint i,s; 
    for (i=0,s=0;i<len;i++){
      if (buffer[i]=='\r' &&  (buffer[i+1]=='\n' || buffer[i+1]=='\0')){
        if (i-s==0) break;
        headers.append(FXString(&buffer[s],i-s));
        s=i+2;
        i++;
        }
      }
    s=i=i+2;
    FXString response(&buffer[s],len-i);
    FXString code=response.section('\n',0);
    FXTRACE((70,"Submit Response: %s\n\n",response.text()));
    if (compare(code,"OK",2)==0) {
      queue.erase(0,submit_npayload);
      clear_submit_payload();
      if (!disconnect) app->addChore(this,ID_NEXT_TASK);
      }
    else if (compare(code,"BADSESSION",10)==0) {
      session.clear();
      clear_submit_payload();
      if (!disconnect) app->addChore(this,ID_NEXT_TASK);
      }
    else {
      set_failed();
      if (!disconnect) app->addTimeout(this,ID_NEXT_TASK,TIME_SEC(SCHEDULE_DELAY));
      }
    }
  else {
    FXTRACE((70,"Submit Response: %s\n\n",buffer));
    set_failed();
    if (!disconnect) app->addTimeout(this,ID_NEXT_TASK,TIME_SEC(SCHEDULE_DELAY));
    }
  } 



long GMAudioScrobbler::onSubmit(FXObject*,FXSelector sel,void*){
  FXTRACE((60,"%s::onSubmit\n",getClassName()));
  FXint ncount;
  FXString msg;
  FXchar buffer[4096]={0};
  switch(FXSELTYPE(sel)){
    case SEL_IO_READ  : 
      ncount=::read(server,buffer,4095);
      process_submit_response(buffer,ncount);
      close_connection();    
      break;
    case SEL_IO_WRITE :
      if (disconnect) {
        close_connection();
        }
      else {
        app->removeInput(server,INPUT_WRITE);
        create_submit_request(msg);
        ncount = ::write(server,(void*)msg.text(),msg.length());
        if (ncount==-1 || ncount!=msg.length()) {
          set_failed();
          close_connection();
          }
        }
      break;
    case SEL_IO_EXCEPT:
      set_failed();
      close_connection(); 
      break;
    }
  return 1;
  }

/******************************************************************************
 *
 *  N O W  P L A Y I N G
 *
 ******************************************************************************/
void GMAudioScrobbler::create_nowplaying_request(FXString & msg) {
  FXTRACE((60,"%s::create_nowplaying_request\n",getClassName()));
  FXString hostname;
  FXString path;
  FXint port;
  FXString nowplaying_payload;  
  nowplaying_payload.clear();
  nowplaying_payload+="s=";
  nowplaying_payload+=myencode(session);
  nowplaying_payload+=GMStringFormat("&a=%s",npt.artist.text());
  nowplaying_payload+=GMStringFormat("&t=%s",npt.title.text());
  nowplaying_payload+=GMStringFormat("&b=%s",npt.album.text());
  nowplaying_payload+=GMStringFormat("&l=%d",npt.duration);
  nowplaying_payload+=GMStringFormat("&n=%d",npt.no);
  nowplaying_payload+="&m";
  npt.artist.clear();
  npt.title.clear();
  npt.album.clear();
  npt.timestamp=0;
  FXTRACE((71,"Payload: %s\n\n",nowplaying_payload.text()));
  parse_url(nowplaying_url,hostname,port,path);
  msg="POST " + path + " HTTP/1.1\r\nHOST:" + hostname + "\r\nContent-Type: application/x-www-form-urlencoded\r\n";
  msg+=GMStringFormat("Content-Length: %d\r\n",nowplaying_payload.length());
  msg+="\r\n";
  msg+=nowplaying_payload;
  }


void GMAudioScrobbler::process_nowplaying_response(const FXchar * buffer,FXint len){
  FXTRACE((60,"%s::process_nowplaying_response\n",getClassName()));
  if (len>0 && compare("HTTP/1.1 200",buffer,12)==0) { 
    FXStringList headers;
    FXint i,s; 
    for (i=0,s=0;i<len;i++){
      if (buffer[i]=='\r' &&  (buffer[i+1]=='\n' || buffer[i+1]=='\0')){
        if (i-s==0) break;
        headers.append(FXString(&buffer[s],i-s));
        s=i+2;
        i++;
        }
      }
    s=i=i+2;
    FXString response(&buffer[s],len-i);
    FXString code=response.section('\n',0);
    FXTRACE((70,"Now Playing Response: %s\n\n",response.text()));
    if (compare(code,"OK",2)==0) {
      clear_submit_payload();
      if (!disconnect) app->addChore(this,ID_NEXT_TASK);
      }
    else if (compare(code,"BADSESSION",10)==0) {
      session.clear();
      clear_submit_payload();
      if (!disconnect) app->addChore(this,ID_NEXT_TASK);
      }
    }
  else {
    FXTRACE((70,"Now Playing Response:\n%s\n\n",buffer));
    }
  } 



long GMAudioScrobbler::onNowPlaying(FXObject*,FXSelector sel,void*){
  FXTRACE((60,"%s::onNowPlaying\n",getClassName()));
  FXint ncount;
  FXString msg;
  FXchar buffer[4096]={0};
  switch(FXSELTYPE(sel)){
    case SEL_IO_READ  : 
      ncount=::read(server,buffer,4095);
      process_nowplaying_response(buffer,ncount);
      close_connection();    
      break;
    case SEL_IO_WRITE :
      if (disconnect) {
        close_connection();
        }
      else {
        app->removeInput(server,INPUT_WRITE);
        create_nowplaying_request(msg);
        ncount = ::write(server,(void*)msg.text(),msg.length());
        if (ncount==-1 || ncount!=msg.length()) {
          set_failed();
          close_connection();
          }
        }
      break;
    case SEL_IO_EXCEPT:
      close_connection(); 
      break;
    }
  return 1;
  }








/******************************************************************************
 *
 *  H A N D S H A K E
 *
 ******************************************************************************/

void GMAudioScrobbler::create_handshake_request(FXString & msg) {
  FXTRACE((60,"%s::create_handshake_request\n",getClassName()));
  FXString token;
  FXlong timestamp = FXThread::time()/1000000000;
	FXString timestamp_text = GMStringVal(timestamp);

  token = password + timestamp_text;
  checksum(token);
          
  msg=GMStringFormat("GET /?hs=true&p=1.2&c="CLIENT_ID"&v="CLIENT_VERSION"&u=%s&t=%s&a=%s HTTP/1.1\r\nHOST:"LASTFM_HOSTNAME"\r\n\r\n",username.text(),timestamp_text.text(),token.text());
  }


void GMAudioScrobbler::process_handshake_response(const FXchar * buffer,FXint len){
  FXTRACE((60,"%s::process_handshake_response\n",getClassName()));
  if (len>0 && compare("HTTP/1.1 200",buffer,12)==0) { 
    FXStringList headers;
    FXint i,s; 
    for (i=0,s=0;i<len;i++){
      if (buffer[i]=='\r' &&  (buffer[i+1]=='\n' || buffer[i+1]=='\0')){
        if (i-s==0) break;
        headers.append(FXString(&buffer[s],i-s));
        s=i+2;
        i++;
        }
      }
    s=i=i+2;
    FXString response(&buffer[s],len-i);
    FXString code=response.section('\n',0);
    if (compare(code,"OK",2)==0) {
      session      = response.section('\n',1);
      nowplaying_url = response.section('\n',2);
      submit_url = response.section('\n',3);
      handshake_failed_timeout=60;
      app->addChore(this,ID_NEXT_TASK);
      //set_permanent_failure(STATE_BANNED);
      }
    else if (compare(code,"BANNED",6)==0){
      set_permanent_failure(STATE_BANNED);
      }
    else if (compare(code,"BADAUTH",7)==0){
      set_permanent_failure(STATE_BADAUTH);
      }
    else if (compare(code,"FAILED",6)==0){
      app->addTimeout(this,ID_NEXT_TASK,TIME_MSEC(handshake_failed_timeout));
      handshake_failed_timeout=FXMIN(MAX_HANDSHAKE_TIMEOUT,handshake_failed_timeout<<1);
      }
    else {
      app->addTimeout(this,ID_NEXT_TASK,TIME_MSEC(handshake_failed_timeout));
      handshake_failed_timeout=FXMIN(MAX_HANDSHAKE_TIMEOUT,handshake_failed_timeout<<1);
      }
    }
  else {
    app->addTimeout(this,ID_NEXT_TASK,TIME_MSEC(handshake_failed_timeout));
    handshake_failed_timeout=FXMIN(MAX_HANDSHAKE_TIMEOUT,handshake_failed_timeout<<1);
    }
  } 

long GMAudioScrobbler::onHandshake(FXObject*,FXSelector sel,void*){
  FXTRACE((60,"%s::onHandshake\n",getClassName()));
  FXint ncount;
  FXString msg;
  FXchar buffer[4096]={0};
  switch(FXSELTYPE(sel)){
    case SEL_IO_READ  :
      ncount=::read(server,buffer,4095);
      if (!disconnect) {
        process_handshake_response(buffer,ncount);
        }
      close_connection();
      break;
    case SEL_IO_WRITE :
      if (disconnect) {
        close_connection();
        }
      else {
        state=STATE_CONNECTED;
        app->removeInput(server,INPUT_WRITE);
        create_handshake_request(msg);
        ncount=::write(server,(void*)msg.text(),msg.length());
        if (ncount==-1) {
          close_connection();
          app->addTimeout(this,ID_NEXT_TASK,TIME_MSEC(handshake_failed_timeout));
          handshake_failed_timeout=FXMIN(MAX_HANDSHAKE_TIMEOUT,handshake_failed_timeout<<1);
          }
        }
      break;
    case SEL_IO_EXCEPT:
      if (disconnect) {
        close_connection();
        }
      else {
        close_connection();
        app->addTimeout(this,ID_NEXT_TASK,TIME_MSEC(handshake_failed_timeout));
        handshake_failed_timeout=FXMIN(MAX_HANDSHAKE_TIMEOUT,handshake_failed_timeout<<1);
        }
      break;
    }
  return 1;
  }
