/*******************************************************************************
*                         Goggles Music Manager                                *
********************************************************************************
*           Copyright (C) 2006-2008 by Sander Jansen. All Rights Reserved      *
*                               ---                                            *
* This program is free software: you can redistribute it and/or modify         *
* it under the terms of the GNU General Public License as published by         *
* the Free Software Foundation, either version 3 of the License, or            *
* (at your option) any later version.                                          *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program.  If not, see http://www.gnu.org/licenses.           *
********************************************************************************/
#include "GMTrackDatabase.h"
#include "GMTag.h"
#include "GMTrackList.h"
#include "GMGenres.h"
#include "GMList.h"
#include "GMSource.h"

/*
*****************************************************
    Goggles Music Manager Database Schema v2008
*****************************************************

    TABLE tracks
        id          INTEGER NOT NULL *
        path        INTEGER
        mrl         TEXT
        title       TEXT NOT NULL
        time        INTEGER
        no          INTEGER
        year        INTEGER
        playcount   INTEGER
        playdate    INTEGER
        importdate  INTEGER
        album       INTEGER NOT NULL
        genre       INTEGER
        rating      INTEGER

    TABLE genres
        id        INTEGER NOT NULL *
        name      TEXT NOT NULL UNIQUE

    TABLE artists
        id        INTEGER NOT NULL *
        name      TEXT NOT NULL UNIQUE

    TABLE albums
        id        INTEGER NOT NULL *
        name      TEXT

(Proposed Extra Data
        year			INTEGER		(to sort albums by year. Derived from track year)

    TABLE album_artist
        album     INTEGER NOT NULL *
        artist    INTEGER NOT NULL

    TABLE playlists
        id        INTEGER NOT NULL *
        name      TEXT

    TABLE playlist_tracks
        playlist  INTEGER
        track     INTEGER
        queue     INTEGER


*/

static FXString last_dir;
static FXint    last_path=-1;



const FXchar create_streams[]="CREATE TABLE streams ( id INTEGER NOT NULL, "
                                                     "url TEXT, "
                                                     "description TEXT, "
                                                     "genre INTEGER, "
                                                     "bitrate INTEGER, "
                                                     "rating INTEGER, "
                                                     "PRIMARY KEY (id) );";

const FXchar create_tracks[]="CREATE TABLE tracks ( id INTEGER NOT NULL, "
                                                   "path INTEGER, "
                                                   "mrl TEXT, "
                                                   "title TEXT NOT NULL, "
                                                   "time INTEGER, "
                                                   "no INTEGER, "
                                                   "year INTEGER, "
                                                   "playcount INTEGER, "
                                                   "playdate INTEGER, "
                                                   "importdate INTEGER, "
                                                   "album INTEGER NOT NULL, "
                                                   "genre INTEGER, "
                                                   "rating INTEGER, "
                                                   "PRIMARY KEY (id) );";

const FXchar create_genres[]="CREATE TABLE genres ( id INTEGER NOT NULL, name TEXT NOT NULL UNIQUE, PRIMARY KEY (id) );";
const FXchar create_albums[]="CREATE TABLE albums ( id INTEGER NOT NULL, name TEXT NOT NULL, PRIMARY KEY (id) );";
const FXchar create_artists[]="CREATE TABLE artists ( id INTEGER NOT NULL, name TEXT NOT NULL UNIQUE, PRIMARY KEY (id) );";
const FXchar create_album_artist[]="CREATE TABLE album_artist ( album INTEGER NOT NULL, artist INTEGER NOT NULL,PRIMARY KEY (album) );";
const FXchar create_playlists[]="CREATE TABLE playlists ( id INTEGER NOT NULL, name TEXT, PRIMARY KEY (id) );";
const FXchar create_playlist_tracks[]="CREATE TABLE playlist_tracks ( playlist INTEGER NOT NULL, track INTEGER NOT NULL, queue INTEGER );";

const FXchar create_directories[]="CREATE TABLE directories ( id INTEGER PRIMARY KEY,parent INTEGER,name TEXT,path TEXT);";


const FXchar create_trigger_diradd[]="CREATE TRIGGER directories_insert AFTER INSERT ON directories "
                                     "BEGIN "
                                     "UPDATE directories SET path = "
                                     "CASE WHEN parent ISNULL THEN '/' "
                                     "ELSE (SELECT path FROM directories WHERE id = new.parent) || parent || '/' "
                                     "END "
                                     "WHERE id = new.id; "
                                     "END;";

// Recursive Trigger not supported by SQLite
//const FXchar create_trigger_dirdel[]="CREATE TRIGGER directories_delete AFTER DELETE ON directories "
//                                     "BEGIN "
//                                     "DELETE FROM directories WHERE id IN "
//                                     "(SELECT id FROM directories WHERE path LIKE old.path || old.id || '/%' ); "
//                                     "END;";


GMTrackDatabase::GMTrackDatabase()  {
  }

GMTrackDatabase::~GMTrackDatabase() {
  }


FXbool GMTrackDatabase::clearTracks(FXbool removeplaylists){

  if (!db.execute("DELETE FROM tracks;"))
    return false;

  if (!db.execute("DELETE FROM artists"))
    return false;

  if (!db.execute("DELETE FROM albums;"))
    return false;

  if (!db.execute("DELETE FROM album_artist;"))
    return false;

  if (!db.execute("DELETE FROM playlist_tracks;"))
    return false;

  if (removeplaylists) {
    if (!db.execute("DELETE FROM playlists;"))
      return false;
    }

  if (!db.execute("DELETE FROM directories;"))
    return false;

  last_path=-1;

  if (!db.execute("DELETE FROM genres WHERE id NOT IN (SELECT DISTINCT(genre) FROM tracks) AND id NOT IN (SELECT DISTINCT(genre) FROM streams);"))
    return false;

  return true;
  }


void GMTrackDatabase::clear(){
  db.clearTables();
  }

FXbool GMTrackDatabase::init(const FXString & database) {
  if (!db.open(database))
    return FALSE;

  if (!setup_tables())
    return FALSE;

  if (!setup_queries())
    return FALSE;

  return true;
  }


void GMTrackDatabase::setup_insert_queries() {
  }


void GMTrackDatabase::clear_insert_queries() {
  }


void GMTrackDatabase::beginDelete() {
  delete_artist.compile(db,"DELETE FROM artists WHERE id == ?;");
  delete_genre.compile(db,"DELETE FROM genres WHERE id == ?;");
  delete_album.compile(db,"DELETE FROM albums WHERE id == ?;");
  delete_track.compile(db,"DELETE FROM tracks WHERE id == ?;");
  delete_track_artist.compile(db,"DELETE FROM tracks WHERE album IN ( SELECT album FROM album_artist WHERE artist == ?);");
  delete_track_album.compile(db,"DELETE FROM tracks WHERE album == ?;");
  delete_albums_artist.compile(db,"DELETE FROM albums WHERE id IN (SELECT album FROM album_artist WHERE artist == ?)");
  delete_artist_album.compile(db,"DELETE FROM album_artist WHERE artist == ?;");
  delete_album_artist.compile(db,"DELETE FROM album_artist WHERE album == ?;");
  delete_track_from_playlist.compile(db,"DELETE FROM playlist_tracks WHERE track == ? AND playlist == ?;");
  delete_track_from_playlists.compile(db,"DELETE FROM playlist_tracks WHERE track == ?;");
  delete_artist_from_playlists.compile(db,"DELETE FROM playlist_tracks WHERE track IN ( SELECT id FROM TRACKS WHERE album IN (SELECT album FROM album_artist WHERE artist == ?));");
  delete_album_from_playlists.compile(db,"DELETE FROM playlist_tracks WHERE track IN (SELECT id FROM tracks WHERE album == ?);");
  query_track_no.compile(db,"SELECT no FROM tracks WHERE id == ?;");
  }

void GMTrackDatabase::endDelete(){
  vacuum();
  delete_artist.clear();
  delete_genre.clear();
  delete_track.clear();
  delete_track_artist.clear();
  delete_track_album.clear();
  delete_album.clear();
  delete_albums_artist.clear();
  delete_artist_album.clear();
  delete_album_artist.clear();
  delete_artist_from_playlists.clear();
  delete_album_from_playlists.clear();
  delete_track_from_playlist.clear();
  delete_track_from_playlists.clear();
	query_track_no.clear();
  }

void GMTrackDatabase::beginEdit() {
  update_track_genre_id.compile(db,"UPDATE tracks SET genre = ? WHERE id == ?;");
  update_mrl.compile(db,"UPDATE tracks SET path = ?, mrl = ? WHERE id == ?;");
  update_artist_album_id.compile(db,"UPDATE album_artist SET artist = ? WHERE album == ?;");
  update_track_title.compile(db,"UPDATE tracks SET title = ? WHERE id == ?;");
  update_track_album_id.compile(db,"UPDATE tracks SET album = ? WHERE id == ?;");
  update_track_genre.compile(db,"UPDATE tracks SET genre = ? WHERE genre == ?;");
  update_track_no.compile(db,"UPDATE tracks SET no = ? WHERE id == ?;");
  update_track_year.compile(db,"UPDATE tracks SET year = ? WHERE id == ?;");
  query_track_no.compile(db,"SELECT no FROM tracks WHERE id == ?;");
  }


void GMTrackDatabase::endEdit(){
  vacuum();
  update_track_genre_id.clear();
  update_mrl.clear();
  update_artist_album_id.clear();
  update_track_title.clear();
  update_track_album_id.clear();
  update_track_genre.clear();
  update_track_no.clear();
  update_track_year.clear();
	query_track_no.clear();
  }







#include "GMPlayerManager.h"

#ifdef DEBUG
namespace FX {
  extern FXlong fxgetticks();
}

static FXlong total_ticks = 0;
static FXlong total_has_ticks = 0;
#endif


FXbool GMTrackDatabase::hasTrack(const FXString & dir,const FXString & filename,FXint & trackid) {
  FXint pathid=-1;
  trackid=-1;
#ifdef DEBUG
  FXlong start=fxgetticks();
#endif
  if (last_dir==dir) {
    if (last_path!=-1) {
      query_filename.setParameter(0,last_path);
      query_filename.setParameter(1,filename);
      query_filename.execute();
      query_filename.getResult(0,trackid);
      query_filename.reset();
      }
    }
  else if (findPath(dir,pathid)){
    query_filename.setParameter(0,pathid);
    query_filename.setParameter(1,filename);
    query_filename.execute();
    query_filename.getResult(0,trackid);
    query_filename.reset();
    last_path=pathid;
    last_dir=dir;
    }
  else {
    last_path=-1;
    last_dir=dir;
    }
#ifdef DEBUG
  FXlong end=fxgetticks();
  total_has_ticks+=(end-start);
#endif

  if (trackid==-1) {
      return false;
      }

  return true;
  }

FXbool GMTrackDatabase::hasTrack(const FXString & filename,FXint & trackid) {
  return hasTrack(FXPath::directory(filename),FXPath::name(filename),trackid);
  }


FXbool GMTrackDatabase::findPath(const FXString & path,FXint & id) {
  if (!FXPath::isTopDirectory(path)) {
    FXString parent = FXPath::upLevel(path);
    if (!findPath(parent,id)) return false;
    FXString name   = FXPath::name(path);
    FXint    nid    = -1;
    query_path_component.setParameter(0,id);
    query_path_component.setParameter(1,name);
    query_path_component.execute();
    query_path_component.getResult(0,nid);
    query_path_component.reset();
    if (nid==-1) return false;
    id=nid;
    }
  else {
    id=-1;
    query_root_component.setParameter(0,path);
    query_root_component.execute();
    query_root_component.getResult(0,id);
    query_root_component.reset();
    if (id==-1) return false;
    }
  return true;
  }

void GMTrackDatabase::addPath(const FXString & path,FXint & id) {
  if (!FXPath::isTopDirectory(path)) {
    FXString parent = FXPath::upLevel(path);
    FXString name   = FXPath::name(path);
    FXint    nid    = -1;

    addPath(parent,id);

    query_path_component.setParameter(0,id);
    query_path_component.setParameter(1,name);
    query_path_component.execute();
    query_path_component.getResult(0,nid);
    query_path_component.reset();
    if (nid==-1) {
      insert_path_component.setParameter(0,id);
      insert_path_component.setParameter(1,name);
      insert_path_component.execute();
      insert_path_component.reset();
      id = db.lastInsertRow();
      }
  else {
      id = nid;
      }
    }
  else {
    id=-1;
    query_root_component.setParameter(0,path);
    query_root_component.execute();
    query_root_component.getResult(0,id);
    query_root_component.reset();
    if (id==-1) {
      insert_root_component.setParameter(0,path);
      insert_root_component.execute();
      insert_root_component.reset();
      id = db.lastInsertRow();
      }
    }
  }



FXbool GMTrackDatabase::insertTrack(const GMTrack & track,FXint & result){
  FXint artist=-1;
  FXint album=-1;
  FXint genre=0;
  FXint path=-1;
  result=0;

#ifdef DEBUG
  FXlong start = fxgetticks();
#endif

  if (!queryArtist(artist,track.artist,true)) return FALSE;
  if (!queryAlbum(album,track.album,artist,true)) return FALSE;
  if (!track.genre.empty() && !queryGenre(genre,track.genre,true)) return FALSE;

  static int cached=0;
  static int newentry=0;

  FXString dir=FXPath::directory(track.mrl);

  if (last_path>=0 && last_dir==dir){
    path=last_path;
    cached++;
    }
  else {
    addPath(dir,path);
    last_path=path;
    last_dir=dir;
    newentry++;
    }

  /// Needed for hasTrack
  try {
    insert_track.setParameter(0,path);

    if (path>=0)
      insert_track.setParameter(1,FXPath::name(track.mrl));
    else
      insert_track.setParameter(1,track.mrl);

    insert_track.setParameter(2,track.title);
    insert_track.setParameter(3,track.time);
    insert_track.setParameter(4,track.no);
    insert_track.setParameter(5,track.year);
    insert_track.setParameter(6,FXThread::time());
    insert_track.setParameter(7,album);
    insert_track.setParameter(8,genre);
    insert_track.execute();
    result = db.lastInsertRow();
//    if (result>0 && trackdict_updated) {
//      trackdict.insert(track.mrl.text(),(void*)(FXival)result);
//      }
    insert_track.reset();
    }
  catch (FXExecuteException & e){
    fxmessage("Exception: %s\n",e.what());
    result=0;
    return FALSE;
    }

#ifdef DEBUG
  FXlong end = fxgetticks();
  total_ticks += (end-start);
#endif
  return true;
  }


FXbool GMTrackDatabase::insertStream(const FXString & url,const FXString & description,const FXString & genre){
  FXString query;
  FXint genreid=0;
  if (!genre.empty() && !queryGenre(genreid,genre,true)) return false;
  query.format("INSERT INTO streams VALUES (NULL,\"%s\",\"%s\",%d,0,0);",url.text(),description.text(),genreid);
  return db.execute(query);
  }


/// Insert Playlist into database
FXbool GMTrackDatabase::insertPlaylist(const FXString & name,FXint & id) {
  try {
    GMQuery query(&db,"INSERT INTO playlists VALUES ( NULL, ? );");
    query.setParameter(0,name);
    query.execute();
    id = db.lastInsertRow();
    query.reset();
    }
  catch (FXCompileException & e){
    return false;
    }
  catch (FXExecuteException & e){
    return false;
    }
  return true;
  }



/// Insert Track in Playlist
FXbool GMTrackDatabase::insertTrackInPlaylist(FXint playlist,FXint & track) {
  FXint queue=0;
  try {
    GMQuery queu_query(&db,"SELECT MAX(queue) FROM playlist_tracks WHERE playlist == ?;");
    queu_query.setParameter(0,playlist);
    queu_query.execute();
    queu_query.getResult(0,queue);
    queu_query.reset();

    // Increment
    queue++;

    GMQuery insert_query(&db,"INSERT INTO playlist_tracks VALUES ( ? , ? , ?);");
    insert_query.setParameter(0,playlist);
    insert_query.setParameter(1,track);
    insert_query.setParameter(2,queue);
    insert_query.execute();

    }
  catch (FXCompileException & e){
    fxmessage("Compile Exception: %s\n",e.what());
    return false;
    }
  catch (FXExecuteException & e){
    fxmessage("Execute Exception: %s\n",e.what());
    return false;
    }
  return true;
  }


/// Insert Track in Playlist
FXbool GMTrackDatabase::insertTrackInPlaylist(FXint playlist,FXIntList & tracks) {
  FXint queue=0;
  try {
    query_playlist_queue.setParameter(0,playlist);
    query_playlist_queue.execute();
    query_playlist_queue.getResult(0,queue);
    query_playlist_queue.reset();

    // Increment
    queue++;

    for (int i=0;i<tracks.no();i++){
      insert_track_playlist.setParameter(0,playlist);
      insert_track_playlist.setParameter(1,tracks[i]);
      insert_track_playlist.setParameter(2,queue++);
      insert_track_playlist.execute();
      insert_track_playlist.reset();
      }

    }
  catch (FXCompileException & e){
    fxmessage("Compile Exception: %s\n",e.what());
    return false;
    }
  catch (FXExecuteException & e){
    fxmessage("Execute Exception: %s\n",e.what());
    return false;
    }
  return true;
  }


/// Insert Track in Playlist
FXbool GMTrackDatabase::updateTrackPlaylists(FXint playlist,FXIntList & tracks) {
  FXint queue=1;
  try {
    GMQuery update_queue(&db,"UPDATE playlist_tracks SET queue = ? WHERE playlist == ? AND track == ?;");
    for (int i=0;i<tracks.no();i++){
      update_queue.setParameter(0,queue++);
      update_queue.setParameter(1,playlist);
      update_queue.setParameter(2,tracks[i]);
      update_queue.execute();
      update_queue.reset();
      }
    }
  catch (FXCompileException & e){
    fxmessage("Compile Exception: %s\n",e.what());
    return false;
    }
  catch (FXExecuteException & e){
    fxmessage("Execute Exception: %s\n",e.what());
    return false;
    }
  return true;
  }


/// List Playlists
FXbool GMTrackDatabase::listPlaylists(FXIntList & ids){
  FXint id=0;
  try {
    GMQuery query(&db,"SELECT id FROM playlists;");
    while(query.execute()){
      query.getResult(0,id);
      ids.append(id);
      }
    query.reset();
    }
  catch (FXCompileException & e){
    fxmessage("Compile Exception: %s\n",e.what());
    ids.clear();
    return false;
    }
  catch (FXExecuteException & e){
    fxmessage("Execute Exception: %s\n",e.what());
    ids.clear();
    return false;
    }
  return true;
  }


FXbool GMTrackDatabase::setStreamFilename(FXint id,const FXString & filename){
  FXString query;
  query.format("UPDATE streams SET url = \"%s\" WHERE id==%d;",filename.text(),id);
  db.execute(query);
  return true;
  }

FXbool GMTrackDatabase::setStreamDescription(FXint id,const FXString & description){
  FXString query;
  query.format("UPDATE streams SET description = \"%s\" WHERE id==%d;",description.text(),id);
  db.execute(query);
  return true;
  }

FXbool GMTrackDatabase::setStreamBitrate(FXint id,FXint rate){
  FXString query;
  query.format("UPDATE streams SET bitrate = \"%d\" WHERE id==%d;",rate,id);
  db.execute(query);
  return true;
  }


FXbool GMTrackDatabase::setStreamGenre(FXint id,const FXString & name){
  FXint genreid=-1;
  FXString query;
  try {

    /// Query for existing genre
    query_genre.setParameter(0,name);
    query_genre.execute();
    query_genre.getResult(0,genreid);
    query_genre.reset();

    if (genreid==-1) {
      /// Insert Genre
      insert_genre.setParameter(0,name);
      insert_genre.execute();
      genreid = db.lastInsertRow();
      insert_genre.reset();
      }

    query.format("UPDATE streams SET genre = %d WHERE id==%d;",genreid,id);
    db.execute(query);
    }
  catch (FXExecuteException &){
    return false;
    }
  return true;
  }




FXbool GMTrackDatabase::removeStream(FXint track) {
  FXString query;
  query.format("DELETE FROM streams WHERE id = \"%d\"",track);
  db.execute(query);
  return true;
  }


void GMTrackDatabase::setTrackFilename(FXint id,const FXString & filename){
  FXint path=-1;
  FXString dir=FXPath::directory(filename);
  addPath(dir,path);
  last_path=path;
  last_dir=dir;
  try {
    update_mrl.setParameter(0,path);
    if (path>=0)
      update_mrl.setParameter(1,FXPath::name(filename));
    else
      update_mrl.setParameter(1,filename);
    update_mrl.setParameter(2,id);
    update_mrl.execute();
    update_mrl.reset();
    }
  catch (FXExecuteException & e){
    }
  }

FXString GMTrackDatabase::getTrackFilename(FXint id){
  FXString filename;
  FXString path;
  FXint    pathid;
  try {
    query_track_filename.setParameter(0,id);
    query_track_filename.execute();
    query_track_filename.getResult(0,filename);
    query_track_filename.getResult(1,pathid);
    query_track_filename.reset();

    query_path.setParameter(0,pathid);
    while(query_path.execute()) {
        if (!path.empty() && path!="/") path+=PATHSEPSTRING;
        path+=query_path.getResult(0);
        }
    query_path.reset();
    }
  catch (FXExecuteException & e){
    return FXString::null;
    }

  if (path.empty())
      return filename;
  else
      return path + PATHSEPSTRING + filename;
  }


FXint GMTrackDatabase::getNumTracks() {
  FXint total=0;
  try {
    GMQuery query(&db,"SELECT COUNT(*) FROM tracks;");
    query.execute();
    query.getResult(0,total);
    query.reset();
    }
  catch (FXCompileException & e){
    return 0;
    }
  catch (FXExecuteException & e){
    return 0;
    }
  return total;
  }

FXint GMTrackDatabase::getTotalTime() {
  FXint total;
  try {
    GMQuery query(&db,"SELECT SUM(time) FROM tracks;");
    query.execute();
    query.getResult(0,total);
    query.reset();
    }
  catch (FXCompileException & e){
    return 0;
    }
  catch (FXExecuteException & e){
    return 0;
    }
  return total;
  }


FXint GMTrackDatabase::getNumArtists() {
  FXint total;
  try {
    GMQuery query(&db,"SELECT COUNT(id) FROM artists;");
    query.execute();
    query.getResult(0,total);
    query.reset();
    }
  catch (FXCompileException & e){
    return 0;
    }
  catch (FXExecuteException & e){
    return 0;
    }
  return total;
  }


FXint GMTrackDatabase::getNumAlbums() {
  FXint total;
  try {
    GMQuery query(&db,"SELECT COUNT(id) FROM albums;");
    query.execute();
    query.getResult(0,total);
    query.reset();
    }
  catch (FXCompileException & e){
    return 0;
    }
  catch (FXExecuteException & e){
    return 0;
    }
  return total;
  }


FXbool GMTrackDatabase::trackInPlaylist(FXint track,FXint playlist) {
  FXint total;
  try {
    GMQuery query(&db,"SELECT COUNT(queue) FROM playlist_tracks WHERE track == ? AND playlist == ? ");
    query.setParameter(0,track);
    query.setParameter(1,playlist);
    query.execute();
    query.getResult(0,total);
    query.reset();
    }
  catch (FXCompileException & e){
    return false;
    }
  catch (FXExecuteException & e){
    return false;
    }
  return (total>0);
  }

FXint GMTrackDatabase::getAlbumPath(FXint id) {
  FXint path;
  try {
    GMQuery query(&db,"SELECT path FROM tracks WHERE album == ? LIMIT(1);");
    query.setParameter(0,id);
    query.execute();
    query.getResult(0,path);
    query.reset();
    }
  catch (FXCompileException & e){
    return -1;
    }
  catch (FXExecuteException & e){
    return -1;
    }
  return path;
  }

FXString GMTrackDatabase::getPathName(FXint id) {
  FXString path;
  try {
    query_path.setParameter(0,id);
    while(query_path.execute()) {
        if (!path.empty() && path!="/") path+=PATHSEPSTRING;
        path+=query_path.getResult(0);
        }
    query_path.reset();
    }
  catch (FXExecuteException & e){
    return FXString::null;
    }
  return path;
  }

FXString GMTrackDatabase::getAlbumName(FXint id) {
  FXString result;
  queryAlbum(result,id);
  return result;
  }

FXString GMTrackDatabase::getAlbumArtist(FXint id) {
  FXString result;
  queryArtist(result,id);
  return result;
  }

FXString GMTrackDatabase::getArtistName(FXint id){
  FXString result;
  try {
    query_artist_name.setParameter(0,id);
    query_artist_name.execute();
    result = query_artist_name.getResult(0);
    query_artist_name.reset();
    }
  catch (FXExecuteException &){
    return FXString::null;
    }
  return result;
  }



FXString GMTrackDatabase::getGenreName(FXint id){
  FXString name;
  try {
    GMQuery query(&db,"SELECT name FROM genres WHERE id == ?;");
    query.setParameter(0,id);
    query.execute();
    query.getResult(0,name);
    query.reset();
    }
  catch (FXCompileException & e){
    return FXString::null;
    }
  catch (FXExecuteException & e){
    return FXString::null;
    }
  return name;
  }

FXint GMTrackDatabase::getTrackAlbum(FXint id){
  FXint result;
  queryAlbum(result,id);
  return result;
  }

/*
FXint GMTrackDatabase::hasTrack(const FXString & mrl){
  FXString filename;
  FXint id;
  if (!trackdict_updated) {
    trackdict.clear();
    try {
      while(list_all_tracks.execute()){
        list_all_tracks.getResult(0,id);
        list_all_tracks.getResult(1,filename);
        trackdict.insert(filename.text(),(void*)(FXival)id);
        }
      list_all_tracks.reset();
      }
    catch (FXExecuteException & e){
      return 0;
      }
    trackdict_updated=true;
    }
  void * data = trackdict.find(mrl.text());
  if (data==NULL)
    return 0;
  else
    return (FXint)(FXival) data;
  }
*/

FXbool GMTrackDatabase::getStream(FXint id,GMStream & stream){
  FXString query;
  query.format("SELECT url,description,genres.name,bitrate FROM streams, genres WHERE genres.id == streams.genre AND streams.id == %d;",id);
  GMQuery get_stream;
  try {
    get_stream.compile(db,query);
    get_stream.execute();
    stream.url=get_stream.getResult(0);
    stream.description=get_stream.getResult(1);
    stream.genre=get_stream.getResult(2);
    get_stream.getResult(3,stream.bitrate);
    get_stream.reset();
    }
  catch (FXCompileException &){
    return false;
    }
  catch (FXExecuteException &){
    return false;
    }
  return true;
  }



FXbool GMTrackDatabase::getTrack(FXint id,GMTrack & track){
  FXString path,filename;
  FXint pathid;
  try {

    query_track.setParameter(0,id);
    query_track.execute();
    query_track.getResult(0,pathid);
    query_track.getResult(1,filename);
    query_track.getResult(2,track.album);
    query_track.getResult(3,track.artist);
    query_track.getResult(4,track.genre);
    query_track.getResult(5,track.title);
    query_track.getResult(6,track.time);
    query_track.getResult(7,track.no);
    query_track.getResult(8,track.year);
    query_track.reset();

    if (pathid>=0) {
      query_path.setParameter(0,pathid);
      while(query_path.execute()) {
        if (!path.empty() && path!="/") path+=PATHSEPSTRING;
        path+=query_path.getResult(0);
        }
      query_path.reset();
      }
    if (!path.empty())
        track.mrl = path + PATHSEPSTRING + filename;
    else
        track.mrl = filename;


/*





    query_track.setParameter(0,id);
    query_track.execute();
    query_track.getResult(0,album);
    query_track.getResult(1,track.mrl);
    query_track.getResult(2,track.title);
    query_track.getResult(3,track.time);
    query_track.getResult(4,track.no);
    query_track.getResult(5,track.year);
    query_track.reset();

    /// Get Album Name
    if (!queryAlbum(track.album,album)) return FALSE;

    /// Get Artist Name
    if (!queryArtist(track.artist,album)) return FALSE;

    /// Get Genre
    if (!queryGenre(track.genre,id)) return FALSE;
*/
    }
  catch (FXExecuteException & e){
    return FALSE;
    }
  return true;
  }

FXbool GMTrackDatabase::getTrackAssociation(FXint id,FXint & artist,FXint & album){

  if (!queryAlbum(album,id)) return FALSE;

  if (!queryArtist(artist,album)) return FALSE;

  return true;
  }


FXbool GMTrackDatabase::removeGenre(FXint id,FXbool update_tracks) {
  GMQuery remove_genre;
  GMQuery remove_tracks;
  GMQuery remove_albums;
  GMQuery remove_albums_artist;
  GMQuery remove_artist;
  try {
      remove_genre.compile((GMDatabase*)this,"DELETE FROM genres WHERE id == ?;");
      remove_tracks.compile((GMDatabase*)this,"DELETE FROM tracks WHERE genre == ?;");
      remove_albums.compile((GMDatabase*)this,"DELETE FROM albums WHERE id NOT IN ( SELECT album FROM tracks);");
      remove_albums_artist.compile((GMDatabase*)this,"DELETE FROM album_artist WHERE album NOT IN ( SELECT id FROM albums );");
      remove_artist.compile((GMDatabase*)this,"DELETE FROM artists WHERE id NOT IN ( SELECT artist FROM album_artist ); ");

      remove_genre.setParameter(0,id);
      remove_genre.perform();

      if (update_tracks) {

        remove_tracks.setParameter(0,id);
        remove_tracks.perform();

        remove_albums.perform();
        remove_albums_artist.perform();
        remove_artist.perform();

        remove_tracks.reset();
        remove_albums.reset();
        remove_albums_artist.reset();
        remove_artist.reset();
        }
      remove_genre.reset();

    }
  catch (FXCompileException & e){
    return false;
    }
  catch (FXExecuteException & e){
    return false;
    }
  return true;
  }



FXbool GMTrackDatabase::removeArtist(FXint artist) {
  try {
    /// Remove tracks from artist from playlists.
    delete_artist_from_playlists.setParameter(0,artist);
    delete_artist_from_playlists.execute();
    delete_artist_from_playlists.reset();

    /// Remove Tracks
    delete_track_artist.setParameter(0,artist);
    delete_track_artist.execute();
    delete_track_artist.reset();

    /// Remove Album Entries
    delete_albums_artist.setParameter(0,artist);
    delete_albums_artist.execute();
    delete_albums_artist.reset();

    /// Remove album_artist entries
    delete_artist_album.setParameter(0,artist);
    delete_artist_album.execute();
    delete_artist_album.reset();

    /// Remove Artist itself
    delete_artist.setParameter(0,artist);
    delete_artist.execute();
    delete_artist.reset();
    }
  catch (FXExecuteException & e){
    return false;
    }
  return true;
  }

FXbool GMTrackDatabase::removeAlbum(FXint album) {
  try {
    delete_album_from_playlists.setParameter(0,album);
    delete_album_from_playlists.execute();
    delete_album_from_playlists.reset();

    delete_track_album.setParameter(0,album);
    delete_track_album.execute();
    delete_track_album.reset();

    delete_album.setParameter(0,album);
    delete_album.execute();
    delete_album.reset();

    delete_album_artist.setParameter(0,album);
    delete_album_artist.execute();
    delete_album_artist.reset();
    }
  catch (FXExecuteException & e){
    return false;
    }
  return true;
  }


FXbool GMTrackDatabase::removeTrack(FXint track) {
  try {
    delete_track.setParameter(0,track);
    delete_track.execute();
    delete_track.reset();

    delete_track_from_playlists.setParameter(0,track);
    delete_track_from_playlists.execute();
    delete_track_from_playlists.reset();
    }
  catch (FXExecuteException & e){
    return false;
    }
  return true;
  }


/// For each playlist, we reset the queue number from 1 to max tracks.
FXbool GMTrackDatabase::reorderPlaylists(){
  GMQuery count_tracks;
  GMQuery count_playlists;
  GMQuery get_tracks;
  GMQuery update_tracks;
  try {
    FXIntList playlists;
    FXIntList rows;
    FXint num_playlists=0;
    FXint num_tracks=0;
    FXint id;
    FXint j=0;

    count_playlists.compile(db,"SELECT id FROM playlists");
    while(count_playlists.execute()) {
      count_playlists.getResult(0,id);
      num_playlists++;
      playlists.append(id);
      }
    count_playlists.reset();

    if (num_playlists) {
      count_tracks.compile(db,"SELECT COUNT(*) FROM playlist_tracks WHERE playlist == ?;");
      get_tracks.compile(db,"SELECT ROWID FROM playlist_tracks WHERE playlist == ? ORDER BY queue;");
      update_tracks.compile(db,"UPDATE playlist_tracks SET queue = ? WHERE ROWID == ?;");
      for (FXint i=0;i<num_playlists;i++){

        /// count how many tracks we have in the playlist
        count_tracks.setParameter(0,playlists[i]);
        count_tracks.execute();
        count_tracks.getResult(0,num_tracks);
        count_tracks.reset();

        if (num_tracks<=0) continue;

        /// Allocate row and track arrays
        rows.no(num_tracks);

        /// Get Row and Track in queue order
        j=0;
        get_tracks.setParameter(0,playlists[i]);
        while(get_tracks.execute()){
          get_tracks.getResult(0,rows[j]);
          j++;
          }
        get_tracks.reset();

        /// Update Queue Number for each track in playlist
        for (j=0;j<num_tracks;j++) {
          update_tracks.setParameter(0,j+1);
          update_tracks.setParameter(1,rows[j]);
          update_tracks.execute();
          update_tracks.reset();
          }
        }
      }
    }
  catch (FXCompileException & e){
    return false;
    }
  catch (FXExecuteException & e){
    return false;
    }
  return true;
  }

     /// Remove track from playlist with specified queue
//      DELETE FROM playlist_tracks WHERE queue == @queue AND playlist == @playlist
//      UPDATE playlist_tracks SET queue = queue - 1 WHERE queue > @queue AND playlist == @playlist;


/*
FXbool GMTrackDatabase::removeTrackFromPlaylist(FXIntList & tracks,FXint playlist){
  try {
    GMQuery query_remove_tracks(&db,"DELETE FROM playlist_tracks WHERE playlist == ? AND track == ?;");
    db.begin();
    for (FXint i=0;i<tracks.no();i++){
      query_remove_tracks.setParameter(0,playlist);
      query_remove_tracks.setParameter(1,tracks[i]);
      query_remove_tracks.execute();
      query_remove_tracks.reset();
      }
    db.commit();
    }
  catch (FXCompileException & e){
    db.rollback();
    return false;
    }
  catch (FXExecuteException & e){
   db.rollback();
    return false;
    }
  return true;
  }
*/









FXbool GMTrackDatabase::removeTracksFromPlaylist(const FXIntList & queue,FXint playlist){
  FXString query;

  if (queue.no()==1) {

    query.format("DELETE FROM playlist_tracks WHERE playlist == %d AND queue == %d;",playlist,queue[0]);
    if (!db.execute(query))
      return false;

    // Renumber queue following removed queue
    query.format("UPDATE playlist_tracks SET queue = queue - 1 WHERE playlist == %d AND queue > %d;",playlist,queue[0]);
    if (!db.execute(query))
      return false;

    }
  else {
    query.format("DELETE FROM playlist_tracks WHERE playlist == %d AND queue IN ( %d",playlist,queue[0]);
    for (FXint i=1;i<queue.no();i++){
      query+=",";
      query+=GMStringVal(queue[i]);
      }
    query+=");";

    if (!db.execute(query))
      return false;

    if (!reorderPlaylists())
      return false;
    }
  return true;
  }



FXbool GMTrackDatabase::removePlaylist(FXint playlist){
  try {
    GMQuery query_remove_tracks(&db,"DELETE FROM playlist_tracks WHERE playlist == ?;");
    GMQuery query_remove_playlist(&db,"DELETE FROM playlists WHERE id == ?;");

    db.begin();

    query_remove_tracks.setParameter(0,playlist);
    query_remove_tracks.execute();
    query_remove_tracks.reset();

    query_remove_playlist.setParameter(0,playlist);
    query_remove_playlist.execute();
    query_remove_playlist.reset();

    db.commit();
    }
  catch (FXCompileException & e){
    db.rollback();
    return false;
    }
  catch (FXExecuteException & e){
   db.rollback();
    return false;
    }
  return true;
  }



/// Get Playlist Name
FXString GMTrackDatabase::getPlaylistName(FXint playlist) {
  FXString name;
  try {
    GMQuery query(&db,"SELECT name FROM playlists WHERE id == ?;");
    query.setParameter(0,playlist);
    query.execute();
    query.getResult(0,name);
    query.reset();
    }
  catch (FXCompileException & e){
    return FXString::null;
    }
  catch (FXExecuteException & e){
    return FXString::null;
    }
  return name;
  }

/// Set Playlist Name
FXbool GMTrackDatabase::setPlaylistName(FXint playlist,const FXString & name) {
  try {
    GMQuery query(&db,"UPDATE playlists SET name = ? WHERE id == ?;");
    query.setParameter(0,name);
    query.setParameter(1,playlist);
    query.execute();
    query.reset();
    }
  catch (FXCompileException & e){
    return false;
    }
  catch (FXExecuteException & e){
    return false;
    }
  return true;
  }

/// Move Track in playlist
FXbool GMTrackDatabase::moveTrack(FXint playlist,FXint oldq,FXint newq){
  FXString query;
  FXint row=0;
  if (oldq==newq) return true;

  query = "SELECT ROWID FROM playlist_tracks WHERE playlist == " + GMStringVal(playlist) + " AND queue == " + GMStringVal(oldq) + ";";
  if (!db.execute_with_result(query,row))
    return false;

  if (oldq<newq)
    query = "UPDATE playlist_tracks SET queue = queue - 1 WHERE playlist == "+ GMStringVal(playlist) +" AND queue > " +GMStringVal(oldq) + " AND queue <= " + GMStringVal(newq) + ";";
  else
    query = "UPDATE playlist_tracks SET queue = queue + 1 WHERE playlist == "+ GMStringVal(playlist) +" AND queue < " +GMStringVal(oldq) + " AND queue >= " + GMStringVal(newq) + ";";

  if (!db.execute(query))
    return false;

  query = "UPDATE playlist_tracks SET queue = " + GMStringVal(newq) + " WHERE playlist == " + GMStringVal(playlist) + " AND ROWID == " + GMStringVal(row) + ";";
  if (!db.execute(query))
    return false;

  return true;
  }



/// Set the name of a track
FXbool GMTrackDatabase::setTrackTitle(FXint track,const FXString & name){
  FXString filename;
  try {
    update_track_title.setParameter(0,name);
    update_track_title.setParameter(1,track);
    update_track_title.execute();
    update_track_title.reset();
    }
  catch (FXExecuteException & e){
    return FALSE;
    }
  return true;
  }


FXbool GMTrackDatabase::playedTrack(FXint track,FXlong time) {
  try {
    update_played.setParameter(0,time);
    update_played.setParameter(1,track);
    update_played.execute();
    update_played.reset();
    }
  catch (FXExecuteException & e){
    return false;
    }
  return true;
  }


FXbool GMTrackDatabase::setTrackYear(FXint track,FXuint year){
  FXString mrl;
  try {
    update_track_year.setParameter(0,(FXint)year);
    update_track_year.setParameter(1,track);
    update_track_year.execute();
    update_track_year.reset();
    }
  catch (FXExecuteException & e){
     return FALSE;
     }
  return true;
  }

FXbool GMTrackDatabase::setTrackDisc(FXint track,FXushort disc){
  FXuint no=0xFFFFFFFF;
  try {
    query_track_no.setParameter(0,track);
    query_track_no.execute();
    query_track_no.getResult(0,no);
    query_track_no.reset();

    update_track_no.setParameter(0,GMALBUMNO(disc,GMTRACKNO(no)));
    update_track_no.setParameter(1,track);
    update_track_no.execute();
    update_track_no.reset();
    }
  catch (FXExecuteException & e){
     return FALSE;
     }
  return true;
  }


FXbool GMTrackDatabase::setTrackNumber(FXint track,FXushort number){
  FXuint no=0xFFFFFFFF;
  try {
    query_track_no.setParameter(0,track);
    query_track_no.execute();
    query_track_no.getResult(0,no);
    query_track_no.reset();

    update_track_no.setParameter(0,GMALBUMNO(GMDISCNO(no),number));
    update_track_no.setParameter(1,track);
    update_track_no.execute();
    update_track_no.reset();
    }
  catch (FXExecuteException & e){
     return FALSE;
     }
  return true;
  }




FXbool GMTrackDatabase::setTrackDiscNumber(FXint track,FXuint no){
  try {
    update_track_no.setParameter(0,no);
    update_track_no.setParameter(1,track);
    update_track_no.execute();
    update_track_no.reset();
    }
  catch (FXExecuteException & e){
     return FALSE;
     }
  return true;
  }

FXbool GMTrackDatabase::setTrackGenre(FXint track,const FXString & name){
  FXint genre=-1;
  FXString mrl;
  try {

    /// Query for existing genre
    query_genre.setParameter(0,name);
    query_genre.execute();
    query_genre.getResult(0,genre);
    query_genre.reset();

    if (genre==-1) {
      /// Insert Genre
      insert_genre.setParameter(0,name);
      insert_genre.execute();
      genre = db.lastInsertRow();
      insert_genre.reset();
      }

    /// Set Genre
    update_track_genre_id.setParameter(0,genre);
    update_track_genre_id.setParameter(1,track);
    update_track_genre_id.execute();
    update_track_genre_id.reset();
    }
  catch (FXExecuteException & e){
     return FALSE;
     }
  return true;
  }


/// Set Track Album
FXbool GMTrackDatabase::setTrackAlbum(FXint track,const FXString & name){
  FXint current_album=-1;
  FXint artist=-1;
  FXint album=-1;
  FXString mrl;
  try {

    /// Fetch Album of track
    if (!queryAlbum(current_album,track)) return FALSE;

    /// Fetch Artist of track
    if (!queryArtist(artist,current_album)) return FALSE;

    /// Fetch Album Id (new or existing) for given album name.
    if (!queryAlbum(album,name,artist,true)) return FALSE;

    /// Set Album
    update_track_album_id.setParameter(0,album);
    update_track_album_id.setParameter(1,track);
    update_track_album_id.execute();
    update_track_album_id.reset();

    /// Vacuum Album
    ///vacuumAlbum(current_album);
    }
  catch (FXExecuteException & e){
     return FALSE;
     }
  return true;
  }


/// Set Track Artist
FXbool GMTrackDatabase::setTrackArtist(FXint track,const FXString & artist_name,const FXString & album_title){
  FXString album_name;
  FXint current_album=-1;
  FXint artist=-1;
  FXint album=-1;
  try {

    /// Fetch Album of track
    if (!queryAlbum(current_album,track)) return FALSE;

    /// Fetch new artist id
    if (!queryArtist(artist,artist_name,true)) return FALSE;

    /// Fetch new album id
    if (album_title.empty()) {
      if (!queryAlbum(album_name,current_album)) return FALSE;
      }
    else {
      album_name=album_title;
      }

    /// Get Album Id for album_name gor given artist
    if (!queryAlbum(album,album_name,artist,true)) return FALSE;

    /// Set New Album id
    update_track_album_id.setParameter(0,album);
    update_track_album_id.setParameter(1,track);
    update_track_album_id.execute();
    update_track_album_id.reset();

    /// Vacuum Album (and Artist)
    //vacuumAlbum(current_album);
    }
  catch (FXExecuteException & e){
     return FALSE;
     }
  return true;
  }

/// Set Track Artist
FXbool GMTrackDatabase::setTrackArtist(FXint track,const FXString & artist_name){
  FXString album_name;
  FXint current_album=-1;
  FXint artist=-1;
  FXint album=-1;
  try {

    /// Fetch Album of track
    if (!queryAlbum(current_album,track)) return FALSE;

    /// Fetch Album Name
    album_name = getAlbumName(current_album);

    /// Fetch new artist id
    if (!queryArtist(artist,artist_name,true)) return FALSE;

    /// Fetch new album id
    if (!queryAlbum(album,album_name,artist,true)) return FALSE;

    /// Set New Album id
    update_track_album_id.setParameter(0,album);
    update_track_album_id.setParameter(1,track);
    update_track_album_id.execute();
    update_track_album_id.reset();

    /// Vacuum Album (and Artist)
    //vacuumAlbum(current_album);
    }
  catch (FXExecuteException & e){
     return FALSE;
     }
  return true;
  }


FXbool GMTrackDatabase::listGenres(FXComboBox * list,FXbool insert_default){
  register int i=0;
  FXDict genres;
  FXString name;
  FXint id;
  try {
    if (insert_default) {
      for (i=0;default_genres[i]!=NULL;i++){
        genres.insert(default_genres[i],(void*)(FXival)1);
        list->appendItem(default_genres[i]);
        }
      while(list_genres.execute()) {
        list_genres.getResult(0,id);
        list_genres.getResult(1,name);
        if (!genres.find(name.text()))
          list->appendItem(name);
        }
      }
    else {
      while(list_genres.execute()) {
        list_genres.getResult(0,id);
        list_genres.getResult(1,name);
        list->appendItem(name);
        }
      }
    list_genres.reset();
    list->sortItems();
    }
  catch(FXExecuteException & e){
    list->clearItems();
    return FALSE;
    }
  return true;
  }




FXbool GMTrackDatabase::listArtists(FXComboBox * list){
  FXString name;
  FXint id;
  try {
    while(list_artists.execute()){
      list_artists.getResult(0,id);
      list_artists.getResult(1,name);
      list->appendItem(name);
      }
    list_artists.reset();
    list->sortItems();
    }
  catch(FXExecuteException & e){
    list->clearItems();
    return FALSE;
    }
  return true;
  }

FXbool GMTrackDatabase::listAlbums(FXComboBox * list,FXint album){
  FXString name;
  FXint id;
  FXint artist;
  try {
    if (!queryArtist(artist,album)) return FALSE;

    list_albums_artist.setParameter(0,artist);
    while(list_albums_artist.execute()){
      list_albums_artist.getResult(0,id);
      list_albums_artist.getResult(1,name);
      list->appendItem(name);
      }
    list_albums_artist.reset();
    list->sortItems();
    }
  catch(FXExecuteException & e){
    list->clearItems();
    return FALSE;
    }
  return true;
  }


FXbool GMTrackDatabase::listTracks(FXIntList & list,FXint album,FXint genre/*=-1*/,FXint playlist/*=-1*/){
  GMQuery query;
  FXint id;
  try {
    if (genre==-1) {
      if (playlist==-1) {
        list_tracks_from_album.setParameter(0,album);
        while(list_tracks_from_album.execute()){
          list_tracks_from_album.getResult(0,id);
          list.append(id);
          }
        list_tracks_from_album.reset();
        }
      else {
        query.compile((GMDatabase*)this,"SELECT tracks.id FROM tracks,genres,playlist_tracks WHERE playlist_tracks.track == tracks.id AND tracks.album == ? AND playlist_tracks.playlist == ?;");
        query.setParameter(0,album);
        query.setParameter(1,playlist);
        while(query.execute()) {
          query.getResult(0,id);
          list.append(id);
          }
        query.reset();
        }
      }
    else {
      if (playlist==-1) {
        list_tracks_from_album_genre.setParameter(0,album);
        list_tracks_from_album_genre.setParameter(1,genre);
        while(list_tracks_from_album_genre.execute()){
          list_tracks_from_album_genre.getResult(0,id);
          list.append(id);
          }
        list_tracks_from_album_genre.reset();
        }
      else {
        query.compile((GMDatabase*)this,"SELECT tracks.id FROM tracks,genres,playlist_tracks WHERE genres.id == tracks.genre AND playlist_tracks.track == tracks.id AND tracks.album == ? AND tracks.genre == ? AND playlist_tracks.playlist == ?;");
        query.setParameter(0,album);
        query.setParameter(1,genre);
        query.setParameter(2,playlist);
        while(query.execute()) {
          query.getResult(0,id);
          list.append(id);
          }
        query.reset();
        }
      }
    }
  catch(FXCompileException & e){
    fxmessage("compile exception: %s\n",e.what());
    list.clear();
    return false;
    }
  catch(FXExecuteException & e){
    fxmessage("execute exception: %s\n",e.what());
    list.clear();
    return false;
    }
  return true;
  }

FXbool GMTrackDatabase::listTracks(FXIntList & list,FXint genre){
  FXint id;
  try {
    list_tracks_genre_only.setParameter(0,genre);
    while(list_tracks_genre_only.execute()){
      list_tracks_genre_only.getResult(0,id);
      list.append(id);
      }
    list_tracks_genre_only.reset();
    }
  catch(FXExecuteException & e){
    fxwarning("Exception: %s\n",e.what());
    list.clear();
    return false;
    }
  return true;
  }

FXbool GMTrackDatabase::listTracksArtists(FXIntList & list,FXint artist,FXint genre/*=-1*/,FXint playlist/*=-1*/){
  GMQuery query;
  FXint id;
  try {
    if (genre==-1) {
      if (playlist==-1) {
        query.compile((GMDatabase*)this,"SELECT tracks.id FROM tracks,album_artist,artists WHERE tracks.album = album_artist.album AND artists.id == album_artist.artist AND artists.id == ?;");
        query.setParameter(0,artist);
        while(query.execute()) {
          query.getResult(0,id);
          list.append(id);
          }
        query.reset();
        }
      else {
        query.compile((GMDatabase*)this,"SELECT tracks.id FROM tracks,album_artist,artists,playlist_tracks WHERE tracks.album == album_artist.album AND artists.id == album_artist.artist AND playlist_tracks.track == tracks.id AND artists.id == ? AND playlist_tracks.playlist == ? ;");
        query.setParameter(0,artist);
        query.setParameter(1,playlist);
        while(query.execute()) {
          query.getResult(0,id);
          list.append(id);
          }
        query.reset();
        }
      }
    else {
      if (playlist==-1) {
        query.compile((GMDatabase*)this,"SELECT tracks.id FROM tracks,album_artist,artists WHERE tracks.album = album_artist.album AND artists.id == album_artist.artist AND tracks.genre == ? AND artists.id == ?;");
        query.setParameter(0,genre);
        query.setParameter(1,artist);
        while(query.execute()) {
          query.getResult(0,id);
          list.append(id);
          }
        query.reset();
        }
      else {
        query.compile((GMDatabase*)this,"SELECT tracks.id FROM tracks,album_artist,artists,playlist_tracks WHERE tracks.album = album_artist.album AND artists.id == album_artist.artist AND playlist_tracks.track == tracks.id AND tracks.genre == ? AND artists.id == ? AND playlist_tracks.playlist == ?;");
        query.setParameter(0,genre);
        query.setParameter(1,artist);
        query.setParameter(2,playlist);
        while(query.execute()) {
          query.getResult(0,id);
          list.append(id);
          }
        query.reset();
        }
      }
    }
  catch(FXCompileException & e){
    fxmessage("compile: %s\n",e.what());
    list.clear();
    return false;
    }
  catch(FXExecuteException & e){
    fxmessage("execute: %s\n",e.what());
    list.clear();
    return false;
    }
  return true;
  }


FXbool GMTrackDatabase::setup_tables() {
  FXString sql;
  FXint  version = db.getVersion();
  GMQuery list_tables;
  GMQuery list_index;
  try {
    if (version==2008) {

      if (!db.hasTable("streams")) {
        if (!db.execute(create_streams))
          return false;
        }

      if (sqlite3_libversion_number()>=SQVERSION(3,3,0)) {
        db.execute("CREATE INDEX IF NOT EXISTS tracks_album ON tracks(album);");
        db.execute("CREATE INDEX IF NOT EXISTS playlist_tracks_idx ON playlist_tracks(track);");
        }
      else {
        if (!db.hasIndex("tracks_album")) db.execute("CREATE INDEX tracks_album ON tracks(album);");
        if (!db.hasIndex("playlist_tracks_idx")) db.execute("CREATE INDEX playlist_tracks_idx ON playlist_tracks(track);");
        }
      return true;
      }

    db.clear();

    /// Create Tables
    if (!db.execute(create_tracks))
      return FALSE;
    if (!db.execute(create_genres))
      return FALSE;
    if (!db.execute(create_albums))
      return FALSE;
    if (!db.execute(create_artists))
       return FALSE;
    if (!db.execute(create_album_artist))
       return FALSE;
    if (!db.execute(create_playlists))
       return FALSE;
    if (!db.execute(create_playlist_tracks))
       return FALSE;
    if (!db.execute(create_directories))
       return FALSE;
    if (!db.execute(create_trigger_diradd))
       return FALSE;
    if (!db.execute(create_streams))
      return FALSE;


    db.execute("CREATE INDEX tracks_album ON tracks(album);");
    db.execute("CREATE INDEX playlist_tracks_idx ON playlist_tracks(track);");
    db.setVersion(2008);
    }
  catch (FXCompileException & e){
      return false;
      }
  catch (FXExecuteException & e){
      return false;
      }
  return true;
  }

FXbool GMTrackDatabase::setup_queries(){
  try {
    insert_root_component.compile(db,"INSERT INTO directories (id,parent,name) VALUES (NULL,NULL,?);");
    insert_path_component.compile(db,"INSERT INTO directories (id,parent,name) VALUES (NULL,?,?);");
    insert_track.compile(db,"INSERT INTO tracks VALUES ( NULL , ? , ? , ? , ? , ? , ? , 0 , 0 , ? , ? , ?, 0 );");
    insert_genre.compile(db,"INSERT INTO genres VALUES ( NULL , ?  );");
    insert_artist.compile(db,"INSERT INTO artists VALUES ( NULL , ?  );");
    insert_album.compile(db,"INSERT INTO albums VALUES ( NULL , ?  );");
    insert_album_artist.compile(db,"INSERT INTO album_artist VALUES ( ? , ?  );");
    query_filename.compile(db,"SELECT id FROM tracks WHERE path == ? AND mrl == ?;");

    query_root_component.compile(db,"SELECT id FROM directories WHERE parent isnull AND name = ?;");
    query_path_component.compile(db,"SELECT id FROM directories WHERE parent = ? AND name = ?;");


    query_path.compile(db,"SELECT name FROM directories WHERE (select path || id || '/' from directories WHERE id = ? ) like path || id || '/%' order by path;");
    query_track.compile(db,"SELECT path, mrl, albums.name, artists.name, genres.name, title, time, no, year "
                           "FROM tracks, albums, genres, artists, album_artist "
                           "WHERE albums.id == tracks.album "
                             "AND genres.id == tracks.genre "
                             "AND album_artist.album == tracks.album "
                             "AND artists.id == album_artist.artist "
                             "AND tracks.id == ?;");

    query_genre.compile(db,"SELECT id FROM genres WHERE name == ?;");
    query_artist.compile(db,"SELECT id FROM artists WHERE name == ?;");
    query_artist_name.compile(db,"SELECT name FROM artists WHERE id == ?;");
    query_album.compile(db,"SELECT albums.id FROM albums,album_artist WHERE albums.id == album_artist.album AND albums.name == ? AND album_artist.artist == ?;");
    query_album_name.compile(db,"SELECT name FROM albums WHERE id == ?;");
    query_artist_with_album.compile(db,"SELECT artist FROM album_artist WHERE album == ?;");
    query_artist_name_album.compile(db,"SELECT name FROM artists WHERE id == ( SELECT artist FROM album_artist WHERE album == ? );");
    query_album_with_track.compile(db,"SELECT album FROM tracks WHERE id == ?;");
    query_track_filename.compile(db,"SELECT mrl,path FROM tracks WHERE tracks.id == ? ;");
    query_track_genre_id.compile(db,"SELECT genre FROM tracks WHERE id == ?;");














/**** DISABLED
    update_mrl.compile(db,"UPDATE tracks SET mrl = ? WHERE id == ?;");
    update_artist.compile(db,"UPDATE artists SET name = ? WHERE id == ?;");
    update_album.compile(db,"UPDATE albums SET name = ? WHERE id == ?;");
    update_artist_album.compile(db,"UPDATE album_artist SET artist = ? WHERE artist == ?;");
    update_artist_album_id.compile(db,"UPDATE album_artist SET artist = ? WHERE album == ?;");
    update_track_title.compile(db,"UPDATE tracks SET title = ? WHERE id == ?;");
    update_track_album.compile(db,"UPDATE tracks SET album = ? WHERE album == ?;");
    update_track_album_id.compile(db,"UPDATE tracks SET album = ? WHERE id == ?;");
    update_track_genre_id.compile(db,"UPDATE tracks SET genre = ? WHERE id == ?;");
    update_track_genre.compile(db,"UPDATE tracks SET genre = ? WHERE genre == ?;");
    update_track_genre_album.compile(db,"UPDATE tracks SET genre = ? WHERE album == ?;");
    update_track_genre_artist.compile(db,"UPDATE tracks SET genre = ? WHERE album IN ( SELECT album FROM album_artist WHERE artist == ? );");
    update_track_no.compile(db,"UPDATE tracks SET no = ? WHERE id == ?;");
    update_genre.compile(db,"UPDATE genres SET name = ? WHERE id == ?;");
*/
    update_played.compile(db,"UPDATE tracks SET playcount = playcount + 1, playdate = ? WHERE id == ?;");








    list_genres.compile(db,"SELECT id, name FROM genres;");
//    list_genres_playlist.compile(db,"SELECT id,name FROM genres WHERE id IN ( SELECT DISTINCT(genre) FROM tracks,playlist_tracks WHERE id == playlist_tracks.track AND playlist_tracks.playlist == ? );");

    list_artists.compile(db,"SELECT id,name FROM artists;");
//    list_artists_genre.compile(db,"SELECT id, name FROM artists WHERE id IN (SELECT DISTINCT(artist) FROM album_artist WHERE album IN ( SELECT DISTINCT(album) FROM tracks WHERE genre == ? ) );");
//    list_artists_playlist.compile(db,"SELECT DISTINCT(artists.id),artists.name FROM artists,tracks,album_artist,playlist_tracks WHERE album_artist.album == tracks.album AND album_artist.artist == artists.id AND playlist_tracks.track == tracks.id AND playlist_tracks.playlist == ?;");
//    list_artists_playlist.compile(db,"SELECT DISTINCT(artists.id), artists.name FROM artists,album_artist WHERE album_artist.artist == artists.id AND album_artist.album IN (SELECT DISTINCT(album) FROM tracks WHERE id IN (SELECT DISTINCT(track) FROM playlist_tracks WHERE playlist == ?));");

//    list_artists_genre_playlist.compile(db,"SELECT DISTINCT(artists.id), artists.name FROM artists,album_artist WHERE album_artist.artist == artists.id AND album_artist.album IN (SELECT DISTINCT(album) FROM tracks WHERE genre==? AND id IN (SELECT DISTINCT(track) FROM playlist_tracks WHERE playlist==?));");


//    list_artists_genre_playlist.compile(db,"SELECT DISTINCT(artists.id),artists.name FROM artists,tracks,album_artist,playlist_tracks WHERE album_artist.album == tracks.album AND album_artist.artist == artists.id AND playlist_tracks.track == tracks.id AND tracks.genre == ? AND playlist_tracks.playlist == ? ;");



   list_albums_artist.compile(db,"SELECT albums.id, albums.name FROM albums,album_artist WHERE albums.id == album_artist.album AND album_artist.artist == ? ;");

//    list_albums_artist_genre.compile(db,"SELECT DISTINCT(albums.id), albums.name  FROM albums,tracks,album_artist WHERE album_artist.album == tracks.album AND album_artist.album == albums.id AND album_artist.artist == ? AND genre == ? ;");
//    list_albums_artist_playlist.compile(db,"SELECT DISTINCT(albums.id), albums.name FROM tracks,albums,album_artist,playlist_tracks WHERE albums.id == tracks.album AND playlist_tracks.track == tracks.id AND album_artist.album == tracks.album AND album_artist.artist == ? AND playlist_tracks.playlist == ?;");

//    list_albums_artist_genre_playlist.compile(db,"SELECT DISTINCT(albums.id), albums.name FROM tracks,albums,album_artist,playlist_tracks WHERE albums.id == tracks.album AND playlist_tracks.track == tracks.id AND album_artist.album == tracks.album AND album_artist.artist == ? AND tracks.genre == ? AND playlist_tracks.playlist == ?;");

//    list_tracks_album_playlist.compile(db,"SELECT tracks.id,title,time,no,genres.name,playlist_tracks.queue FROM tracks,genres,playlist_tracks WHERE genres.id == tracks.genre AND playlist_tracks.track == tracks.id AND tracks.album == ? AND playlist_tracks.playlist == ?");
//    list_tracks_album_genre_playlist.compile(db,"SELECT id,title,time,no,queue FROM tracks,playlist_tracks WHERE playlist_tracks.track == tracks.id AND album == ? AND genre == ? AND playlist_tracks.playlist == ?");

//    list_tracks.compile(db,"SELECT tracks.id,mrl,title,time,no,genres.name FROM tracks, genres WHERE album == ? AND genres.id == tracks.genre");


    list_tracks_from_album.compile(db,"SELECT id FROM tracks WHERE album == ?");
    list_tracks_from_album_genre.compile(db,"SELECT id FROM tracks WHERE album == ? AND genre == ?;");
//    list_tracks_genre.compile(db,"SELECT id,title,time,no FROM tracks WHERE album == ? AND genre == ?;");
    list_tracks_genre_only.compile(db,"SELECT id FROM tracks WHERE genre == ?;");

/*
    list_all_info.compile(db,"SELECT tracks.id,title,artists.name,albums.name,time,no,genres.name "
                             "FROM artists,album_artist,albums,genres,tracks "
                             "WHERE albums.id == tracks.album AND "
                                   "albums.id == album_artist.album AND "
                                   "album_artist.artist == artists.id AND "
                                   "tracks.genre == genres.id ;");
    list_from_playlist.compile(db,"SELECT tracks.id,title,artists.name,albums.name,time,no,genres.name "
                                  "FROM artists,album_artist,albums,genres,tracks,playlist_tracks "
                                  "WHERE albums.id == tracks.album AND "
                                        "albums.id == album_artist.album AND "
                                        "album_artist.artist == artists.id AND "
                                        "tracks.genre == genres.id AND "
                                        "tracks.id == playlist_tracks.track AND "
                                        "playlist_tracks.playlist == ? "
                                  "ORDER BY playlist_tracks.queue;");
*/


    query_playlist_queue.compile(db,"SELECT MAX(queue) FROM playlist_tracks WHERE playlist == ?;");
//    delete_playlist.compile(db,"DELETE FROM playlist_tracks WHERE playlist == ?;");
    insert_track_playlist.compile(db,"INSERT INTO playlist_tracks VALUES ( ? , ? , ?);");
    }
  catch (FXCompileException & e){
    fxmessage("Compile Exception: %s\n",e.what());
    return FALSE;
    }
  return true;
  }


FXbool GMTrackDatabase::queryGenre(FXint & result,const FXString & name,FXbool insert){
  result=-1;
  try {
    query_genre.setParameter(0,name);
    query_genre.execute();
    query_genre.getResult(0,result);
    query_genre.reset();

    if (result==-1 && insert) {
      insert_genre.setParameter(0,name);
      insert_genre.execute();
      result = db.lastInsertRow();
      insert_genre.reset();
      }
    }
  catch (FXExecuteException & e){
    result=-1;
    return FALSE;
    }
  return true;
  }

/*
FXbool GMTrackDatabase::queryGenre(FXString & result,FXint trackid){
  result=FXString::null;
  try {
    query_track_genre.setParameter(0,trackid);
    query_track_genre.execute();
    query_track_genre.getResult(0,result);
    query_track_genre.reset();
    }
  catch (FXExecuteException & e){
    result=FXString::null;
    return FALSE;
    }
  return true;
  }
*/

FXbool GMTrackDatabase::getTrackGenre(FXint & genre,FXint trackid) {
  genre=-1;
  try {
    query_track_genre_id.setParameter(0,trackid);
    query_track_genre_id.execute();
    query_track_genre_id.getResult(0,genre);
    query_track_genre_id.reset();
    }
  catch (FXExecuteException & e){
    genre=-1;
    return FALSE;
    }
  return true;
  }



FXbool GMTrackDatabase::queryArtist(FXint & result,const FXString & name,FXbool insert){
  result=-1;
  try {
    query_artist.setParameter(0,name);
    query_artist.execute();
    query_artist.getResult(0,result);
    query_artist.reset();

    if (result==-1 && insert) {
      insert_artist.setParameter(0,name);
      insert_artist.execute();
      result = db.lastInsertRow();
      insert_artist.reset();
      }
    }
  catch (FXExecuteException & e){
    result=-1;
    return FALSE;
    }
  return true;
  }


FXbool GMTrackDatabase::queryArtist(FXint & result,FXint album){
  result=-1;
  try {
    query_artist_with_album.setParameter(0,album);
    query_artist_with_album.execute();
    query_artist_with_album.getResult(0,result);
    query_artist_with_album.reset();
    }
  catch (FXExecuteException & e){
    result=-1;
    return FALSE;
    }
  return true;
  }


FXbool GMTrackDatabase::queryArtist(FXString & result,FXint album){
  result=FXString::null;
  try {
    query_artist_name_album.setParameter(0,album);
    query_artist_name_album.execute();
    query_artist_name_album.getResult(0,result);
    query_artist_name_album.reset();
    }
  catch (FXExecuteException & e){
    result=FXString::null;
    return FALSE;
    }
  return true;
  }


FXbool GMTrackDatabase::queryAlbum(FXint & result,FXint track){
  result=-1;
  try {
    query_album_with_track.setParameter(0,track);
    query_album_with_track.execute();
    query_album_with_track.getResult(0,result);
    query_album_with_track.reset();
    }
  catch (FXExecuteException & e){
    result=-1;
    return FALSE;
    }
  return true;
  }

FXbool GMTrackDatabase::queryAlbum(FXint & result,const FXString & name,FXint artist,FXbool insert){
  result=-1;
  try {
    query_album.setParameter(0,name);
    query_album.setParameter(1,artist);
    query_album.execute();
    query_album.getResult(0,result);
    query_album.reset();

    if (result==-1 && insert) {
      insert_album.setParameter(0,name);
      insert_album.execute();
      result = db.lastInsertRow();
      insert_album.reset();

      insert_album_artist.setParameter(0,result);
      insert_album_artist.setParameter(1,artist);
      insert_album_artist.execute();
      insert_album_artist.reset();
      }

    }
  catch (FXExecuteException & e){
    fxmessage("Exception: %s\n",e.what());
    result=-1;
    return FALSE;
    }
  return true;
  }


FXbool GMTrackDatabase::queryAlbum(FXString & name,FXint album){
  name=FXString::null;
  try {
    query_album_name.setParameter(0,album);
    query_album_name.execute();
    query_album_name.getResult(0,name);
    query_album_name.reset();
    }
  catch (FXExecuteException & e){
    name=FXString::null;
    return FALSE;
    }
  return true;
  }






FXbool GMTrackDatabase::vacuum() {

#ifdef DEBUG
  FXlong start = fxgetticks();
#endif

  /// Remove empty playlists ? FIXME: may be not...
  /// if (!db.execute("DELETE FROM playlists WHERE id NOT IN (SELECT DISTINCT(playlist) FROM playlist_tracks);"))
  //  return false;

  /// Reorder Playlists
  if (!reorderPlaylists())
    return false;

  /// Remove empty albums
  if (!db.execute("DELETE FROM albums WHERE id NOT IN (SELECT DISTINCT(album) FROM tracks);"))
    return false;

  /// Remove invalid albums from album_artist
  if (!db.execute("DELETE FROM album_artist WHERE album NOT IN (SELECT id FROM albums);"))
    return false;

  /// Remove unused artists
  if (!db.execute("DELETE FROM artists WHERE id NOT IN (SELECT DISTINCT(artist) FROM album_artist);"))
    return false;

  /// Remove unused genres
  if (!db.execute("DELETE FROM genres WHERE id NOT IN (SELECT DISTINCT(genre) FROM tracks) AND id NOT IN (SELECT DISTINCT(genre) FROM streams);"))
    return false;

  /// Remove empty leafs
  do {
    if (!db.execute("DELETE FROM directories WHERE id NOT IN (SELECT DISTINCT(parent) FROM directories) AND id NOT IN (SELECT DISTINCT(path) FROM tracks);"))
      return false;
    }
  while (db.changes()>0);
  last_path=-1;


#ifdef DEBUG
  FXlong end = fxgetticks();
  fxmessage("Vacuum Database in %ld ticks\n",end-start);
#endif

  return true;
  }


FXbool GMTrackDatabase::exportList(const FXString & filename,FXint playlist,FXuint filetype){
  const FXchar * c_artist;
  const FXchar * c_album;
  const FXchar * c_title;
  const FXchar * c_genre;
  FXint time;
  FXint no;
  FXint id;
  FXint cnt=1;
	FXint year;
  FILE * fp = NULL;
  FXString query;
  GMQuery list;
  try {

    query = "SELECT tracks.id,title,artists.name,albums.name,genres.name,no,time,year "
            "FROM tracks, genres, albums, artists, album_artist";

    if (playlist>=0)
      query+=", playlist_tracks";

    query+=" WHERE genres.id == tracks.genre AND "
                  "album_artist.artist == artists.id AND "
                  "album_artist.album == albums.id AND "
                  "tracks.album == albums.id";

    if (playlist>=0) {
      query+=" AND playlist_tracks.track == tracks.id";
      query+=" AND playlist_tracks.playlist == "+GMStringVal(playlist);
      query+=" ORDER BY playlist_tracks.queue; ";
      }
    else {
      query+=";";
      }

    list.compile(db,query);

    FILE * fp = fopen(filename.text(),"w");
    if (!fp) return false;

    if (filetype==PLAYLIST_XSPF) {
      fprintf(fp,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
      fprintf(fp,"<playlist version=\"1\" xmlns=\"http://xspf.org/ns/0/\">\n");
      fprintf(fp,"\t<trackList>\n");
      }
    else if (filetype==PLAYLIST_M3U_EXTENDED) {
      fprintf(fp,"#EXTM3U\n");
      }
    else if (filetype==PLAYLIST_PLS) {
      fprintf(fp,"[playlist]\n");
      }
    else if (filetype==PLAYLIST_CSV) {
      fprintf(fp,"No,Title,Album,Artist,Genre,Duration,Year,Filename\n");
      }

    while(list.execute()){
      list.getResult(0,id);
      c_title = list.getResult(1);
      c_artist = list.getResult(2);
      c_album = list.getResult(3);
      c_genre = list.getResult(4);
      list.getResult(5,no);
      list.getResult(6,time);
      list.getResult(7,year);
      if (filetype==PLAYLIST_XSPF) {
        fprintf(fp,"\t\t<track>\n");
        fprintf(fp,"\t\t\t<location>%s</location>\n",FXURL::fileToURL(getTrackFilename(id)).text());
        fprintf(fp,"\t\t\t<creator>%s</creator>\n",c_artist);
        fprintf(fp,"\t\t\t<album>%s</album>\n",c_album);
        fprintf(fp,"\t\t\t<title>%s</title>\n",c_title);
        fprintf(fp,"\t\t\t<duration>%d</duration>\n",time*1000);
        fprintf(fp,"\t\t\t<trackNum>%u</trackNum>\n",no);
        fprintf(fp,"\t\t</track>\n");
        }
      else if (filetype==PLAYLIST_M3U_EXTENDED) {
        fprintf(fp,"#EXTINF:%d,%s - %s\n",time,c_artist,c_title);
        fprintf(fp,"%s\n",getTrackFilename(id).text());
        }
      else if (filetype==PLAYLIST_M3U) {
        fprintf(fp,"%s\n",getTrackFilename(id).text());
        }
      else if (filetype==PLAYLIST_PLS) {
        fprintf(fp,"File%d=%s\n",cnt,getTrackFilename(id).text());
        fprintf(fp,"Title%d=%s - %s\n",cnt,c_artist,c_title);
        fprintf(fp,"Length%d=%d\n",cnt,time);
        cnt++;
        }
      else if (filetype==PLAYLIST_CSV){
        fprintf(fp,"%d,\"%s\",\"%s\",\"%s\",\"%s\",%d,%d,\"%s\"\n",no,c_title,c_album,c_artist,c_genre,time,year,getTrackFilename(id).text());
        }
      }
    list.reset();
    if (filetype==PLAYLIST_XSPF) {
      fprintf(fp,"\t</trackList>\n");
      fprintf(fp,"</playlist>");
      }
    else if (filetype==PLAYLIST_PLS){
      fprintf(fp,"Version=2\n");
      fprintf(fp,"NumberOfEntries=%d",cnt-1);
      }
    fclose(fp);
    fp=NULL;
    }
  catch (FXCompileException & e){
    if (fp) fclose(fp);
    return false;
    }
  catch (FXExecuteException & e){
    if (fp) fclose(fp);
    return false;
    }
  return true;
  }

